# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================

from dwave_networkx.generators.pegasus import pegasus_graph
import networkx as nx

from minorminer.busclique import busgraph_cache


__all__ = ['find_clique_embedding', 'find_biclique_embedding']

def _get_target_graph(m=None, target_graph=None):
    if target_graph is None:
        if m is None:
            raise TypeError("m and target_graph cannot both be None.")
        target_graph = pegasus_graph(m)
    return target_graph


@nx.utils.decorators.nodes_or_number(0)
def find_clique_embedding(k, m=None, target_graph=None):
    """Find an embedding for a clique in a Pegasus graph.

    Given a clique (fully connected graph) and target Pegasus graph, attempts
    to find an embedding by transforming the Pegasus graph into a :math:`K_{2,2}`
    Chimera graph and then applying a Chimera clique-finding algorithm. Results
    are converted back to Pegasus coordinates.

    Args:
        k (int/iterable/:obj:`networkx.Graph`): A complete graph to embed,
            formatted as a number of nodes, node labels, or a NetworkX graph.
        m (int): Number of tiles in a row of a square Pegasus graph. Required to
            generate an m-by-m Pegasus graph when `target_graph` is None.
        target_graph (:obj:`networkx.Graph`): A Pegasus graph. Required when `m`
            is None.

    Returns:
        dict: An embedding as a dict, where keys represent the clique's nodes and
        values, formatted as lists, represent chains of pegasus coordinates.

    Examples:
        This example finds an embedding for a :math:`K_3` complete graph in a
        2-by-2 Pegaus graph.

        >>> from dwave.embedding.pegasus import find_clique_embedding
        ...
        >>> print(find_clique_embedding(3, 2))    # doctest: +SKIP
        {0: [10, 34], 1: [35, 11], 2: [32, 12]}

    """
    _, nodes = k
    g = _get_target_graph(m, target_graph)
    embedding = busgraph_cache(g).find_clique_embedding(nodes)

    if len(embedding) != len(nodes):
        raise ValueError("No clique embedding found")

    return embedding


@nx.utils.decorators.nodes_or_number(0)
@nx.utils.decorators.nodes_or_number(1)
def find_biclique_embedding(a, b, m=None, target_graph=None):
    """Find an embedding for a biclique in a Pegasus graph.

    Given a biclique (a bipartite graph where every vertex in a set in connected
    to all vertices in the other set) and a target :term:`Pegasus` graph size or
    edges, attempts to find an embedding.

    Args:
        a (int/iterable):
            Left shore of the biclique to embed. If a is an integer, generates
            an embedding for a biclique with the left shore of size a labelled
            [0,a-1]. If a is an iterable of nodes, generates an embedding for a
            biclique with the left shore of size len(a) labelled for the given
            nodes.

        b (int/iterable):
            Right shore of the biclique to embed.If b is an integer, generates
            an embedding for a biclique with the right shore of size b labelled
            [0,b-1]. If b is an iterable of nodes, generates an embedding for a
            biclique with the right shore of size len(b) labelled for the given
            nodes.

        m (int): Number of tiles in a row of a square Pegasus graph. Required to
            generate an m-by-m Pegasus graph when `target_graph` is None.

        target_graph (:obj:`networkx.Graph`): A Pegasus graph. Required when `m`
            is None.

    Returns:
        tuple: A 2-tuple containing:

            dict: An embedding mapping the left shore of the biclique to
            the Pegasus lattice.

            dict: An embedding mapping the right shore of the biclique to
            the Pegasus lattice.

    Examples:
        This example finds an embedding for an alphanumerically labeled biclique in a small
        Pegasus graph

        >>> from dwave.embedding.pegasus import find_biclique_embedding
        ...
        >>> left, right = find_biclique_embedding(['a', 'b', 'c'], ['d', 'e'], 2)
        >>> print(left, right)  # doctest: +SKIP
        {'a': [40], 'b': [41], 'c': [42]} {'d': [4], 'e': [5]}

    """
    _a, anodes = a
    _b, bnodes = b

    if isinstance(_a, int) and isinstance(_b, int):
        bnodes = [len(anodes) + x for x in bnodes]

    if set(anodes).intersection(set(bnodes)):
        raise ValueError("a and b overlap")

    g = _get_target_graph(m, target_graph)
    embedding = busgraph_cache(g).find_biclique_embedding(len(anodes), len(bnodes))

    if not embedding:
        raise ValueError("No biclique embedding found")

    return ({x: embedding[anodes.index(x)] for x in anodes},
            {y: embedding[bnodes.index(y) + len(anodes)] for y in bnodes})



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================

try:
    import collections.abc as abc
except ImportError:
    import collections as abc

from dwave_networkx.generators.chimera import chimera_graph, chimera_coordinates
import networkx as nx

from minorminer.busclique import busgraph_cache

__all__ = ['find_clique_embedding',
           'find_biclique_embedding',
           'find_grid_embedding',
           ]


@nx.utils.decorators.nodes_or_number(0)
def find_clique_embedding(k, m=None, n=None, t=None, target_edges=None, target_graph=None):
    """Find an embedding for a clique in a Chimera graph.

    Given the node labels or size of a clique (fully connected graph) and size or
    edges of the target :term:`Chimera` graph, attempts to find an embedding.

    Args:
        k (int/iterable):
            Clique to embed. If k is an integer, generates an embedding for a
            clique of size k labelled [0,k-1]. If k is an iterable of nodes,
            generates an embedding for a clique of size len(k) labelled
            for the given nodes.

        m (int, optional, default=None):
            Number of rows in the Chimera lattice.

        n (int, optional, default=m):
            Number of columns in the Chimera lattice.

        t (int, optional, default 4):
            Size of the shore within each Chimera tile.

        target_edges (iterable[edge]):
            A list of edges in the target Chimera graph. Nodes are labelled as
            returned by :func:`~dwave_networkx.chimera_graph`.

        target_graph (networkx.Graph):
            A Chimera graph constructed by :func:`~dwave_networkx.chimera_graph`.

    Returns:
        dict: An embedding mapping a clique to the Chimera lattice.

    Note:
        Either target_edges or target_graph must be None.  If both are None,
        a graph with perfect yield is assumed from the parameters m, n, t.  If
        target_edges is not None, at least m must not be None.

    Examples:
        The first example finds an embedding for a :math:`K_4` complete graph in a single
        Chimera unit cell. The second for an alphanumerically labeled :math:`K_3`
        graph in 4 unit cells.

        >>> from dwave.embedding.chimera import find_clique_embedding
        ...
        >>> embedding = find_clique_embedding(4, 1, 1)
        >>> embedding  # doctest: +SKIP
        {0: [4, 0], 1: [5, 1], 2: [6, 2], 3: [7, 3]}

        >>> from dwave.embedding.chimera import find_clique_embedding
        ...
        >>> embedding = find_clique_embedding(['a', 'b', 'c'], m=2, n=2, t=4)
        >>> embedding  # doctest: +SKIP
        {'a': [20, 16], 'b': [21, 17], 'c': [22, 18]}

    """
    _, nodes = k
    m, n, t, g = _get_target_graph(m, n, t, target_edges, target_graph)
    embedding = busgraph_cache(g).find_clique_embedding(nodes)

    if len(embedding) != len(nodes):
        raise ValueError("No clique embedding found")

    return embedding


@nx.utils.decorators.nodes_or_number(0)
@nx.utils.decorators.nodes_or_number(1)
def find_biclique_embedding(a, b, m=None, n=None, t=None, target_edges=None, target_graph=None):
    """Find an embedding for a biclique in a Chimera graph.

    Given a biclique (a bipartite graph where every vertex in a set in connected
    to all vertices in the other set) and a target :term:`Chimera` graph size or
    edges, attempts to find an embedding.

    Args:
        a (int/iterable):
            Left shore of the biclique to embed. If a is an integer, generates
            an embedding for a biclique with the left shore of size a labelled
            [0,a-1]. If a is an iterable of nodes, generates an embedding for a
            biclique with the left shore of size len(a) labelled for the given
            nodes.

        b (int/iterable):
            Right shore of the biclique to embed.If b is an integer, generates
            an embedding for a biclique with the right shore of size b labelled
            [0,b-1]. If b is an iterable of nodes, generates an embedding for a
            biclique with the right shore of size len(b) labelled for the given
            nodes.

        m (int, optional, default=None):
            Number of rows in the Chimera lattice.

        n (int, optional, default=m):
            Number of columns in the Chimera lattice.

        t (int, optional, default 4):
            Size of the shore within each Chimera tile.

        target_edges (iterable[edge]):
            A list of edges in the target Chimera graph. Nodes are labelled as
            returned by :func:`~dwave_networkx.chimera_graph`.

        target_graph (networkx.Graph):
            A Chimera graph constructed by :func:`~dwave_networkx.chimera_graph`.

    Returns:
        tuple: A 2-tuple containing:

            dict: An embedding mapping the left shore of the biclique to
            the Chimera lattice.

            dict: An embedding mapping the right shore of the biclique to
            the Chimera lattice.

    Note:
        Either target_edges or target_graph must be None.  If both are None,
        a graph with perfect yield is assumed from the parameters m, n, t.  If
        target_edges is not None, at least m must not be None.

    Examples:
        This example finds an embedding for an alphanumerically labeled biclique in a single
        Chimera unit cell.

        >>> from dwave.embedding.chimera import find_biclique_embedding
        ...
        >>> left, right = find_biclique_embedding(['a', 'b', 'c'], ['d', 'e'], 1, 1)
        >>> print(left, right)  # doctest: +SKIP
        {'a': [4], 'b': [5], 'c': [6]} {'d': [0], 'e': [1]}

    """
    _a, anodes = a
    _b, bnodes = b

    if isinstance(_a, int) and isinstance(_b, int):
        bnodes = [len(anodes) + x for x in bnodes]

    if set(anodes).intersection(set(bnodes)):
        raise ValueError("a and b overlap")

    m, n, t, g = _get_target_graph(m, n, t, target_edges, target_graph)
    embedding = busgraph_cache(g).find_biclique_embedding(len(anodes), len(bnodes))

    if not embedding:
        raise ValueError("No biclique embedding found")

    return ({x: embedding[anodes.index(x)] for x in anodes},
            {y: embedding[bnodes.index(y) + len(anodes)] for y in bnodes})


def find_grid_embedding(dim, m, n=None, t=4):
    """Find an embedding for a grid in a Chimera graph.

    Given grid dimensions and a target :term:`Chimera` graph size,
    attempts to find an embedding.

    Args:
        dim (iterable[int]):
            Sizes of each grid dimension. Length can be between 1 and 3.

        m (int):
            Number of rows in the Chimera lattice.

        n (int, optional, default=m):
            Number of columns in the Chimera lattice.

        t (int, optional, default 4):
            Size of the shore within each Chimera tile.

    Returns:
        dict: An embedding mapping a grid to the Chimera lattice.

    Examples:
        This example finds an embedding for a 2x3 grid in a 12x12 lattice of Chimera unit cells.

        >>> from dwave.embedding.chimera import find_grid_embedding
        ...
        >>> embedding = find_grid_embedding([2, 3], m=12, n=12, t=4)
        >>> embedding  # doctest: +SKIP
        {(0, 0): [0, 4],
         (0, 1): [8, 12],
         (0, 2): [16, 20],
         (1, 0): [96, 100],
         (1, 1): [104, 108],
         (1, 2): [112, 116]}

    """

    m, n, t, _ = _get_target_graph(m, n, t, None, _dont_construct)
    indexer = chimera_coordinates(m, n, t)

    dim = list(dim)
    num_dim = len(dim)
    if num_dim == 1:
        def _key(row, col, aisle): return row
        dim.extend([1, 1])
    elif num_dim == 2:
        def _key(row, col, aisle): return row, col
        dim.append(1)
    elif num_dim == 3:
        def _key(row, col, aisle): return row, col, aisle
    else:
        raise ValueError("find_grid_embedding supports between one and three dimensions")

    rows, cols, aisles = dim
    if rows > m or cols > n or aisles > t:
        msg = ("the largest grid that find_grid_embedding can fit in a ({}, {}, {}) Chimera-lattice "
               "is {}x{}x{}; given grid is {}x{}x{}").format(m, n, t, m, n, t, rows, cols, aisles)
        raise ValueError(msg)

    return {_key(row, col, aisle): [indexer.chimera_to_linear((row, col, 0, aisle)),
                                     indexer.chimera_to_linear((row, col, 1, aisle))]
            for row in range(dim[0]) for col in range(dim[1]) for aisle in range(dim[2])}

_dont_construct = object()
def _get_target_graph(m, n=None, t=None, target_edges=None, target_graph=None):
    if target_graph is not None and target_graph is not _dont_construct:
        if target_edges is not None:
            raise ValueError("either target_graph or target_edges must be None")
        m = target_graph.graph['rows']
        n = target_graph.graph['columns']
        t = target_graph.graph['tile']
    else:
        if not isinstance(m, int):
            raise TypeError('Chimera lattice parameter m must be an int and >= 1')
        if m <= 0:
            raise ValueError('Chimera lattice parameter m must be an int and >= 1')

        if n is None:
            n = m
        else:
            if not isinstance(n, int):
                raise TypeError('Chimera lattice parameter n must be an int and >= 1')
            if n <= 0:
                raise ValueError('Chimera lattice parameter n must be an int and >= 1')
        if t is None:
            t = 4
        else:
            if not isinstance(t, int):
                raise TypeError('Chimera lattice parameter t must be an int and >= 1')
            if t <= 0:
                raise ValueError('Chimera lattice parameter t must be an int and >= 1')

        if target_graph is not _dont_construct:
            target_graph = chimera_graph(m, n, t, edge_list = target_edges)

    return m, n, t, target_graph



# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import random
from collections import Counter, abc, defaultdict

import networkx as nx
import dwave_networkx as dnx
import numpy as np
from scipy import spatial

from . import layout


def intersection(S_layout, T_layout, **kwargs):
    """Map each vertex of S to its nearest row/column intersection qubit in T 
    (T must be a D-Wave hardware graph). Note: This will modify ``S_layout``. 

    Args:
        S_layout (:class:`.Layout`):
            A layout for S; i.e. a map from S to R^d.

        T_layout (:class:`.Layout`):
            A layout for T; i.e. a map from T to R^d.

    Returns:
        dict: A mapping from vertices of S (keys) to vertices of T (values).
    
    Examples:
        This example creates a :class:`.Placement` object that stores a mapping 
        computed with :func:`.intersection`, in which the nodes from a source 
        hexagonal lattice graph are mapped to a target Chimera graph.

        >>> import networkx as nx
        >>> import dwave_networkx as dnx
        >>> import minorminer.layout as mml
        ...
        >>> G = nx.hexagonal_lattice_graph(2,2)
        >>> G_layout = mml.Layout(G, mml.p_norm)
        >>> C = dnx.chimera_graph(2,2)
        >>> C_layout = mml.Layout(C, mml.dnx_layout)
        >>> placement = mml.Placement(G_layout, C_layout, placement=mml.intersection)

        ``placement`` may be passed in directly to :func:`minorminer.layout.find_embedding`. 
        
        Alternatively, :func:`.intersection` may be passed in instead, as shown 
        in the example below.

        >>> import networkx as nx
        >>> import dwave_networkx as dnx
        >>> import minorminer.layout as mml
        ...
        >>> G = nx.hexagonal_lattice_graph(2,2)
        >>> C = dnx.chimera_graph(2,2)
        >>> embedding = mml.find_embedding(G, 
        ...                                C, 
        ...                                placement=mml.intersection)

    """
    # Extract the target graph
    T = T_layout.G

    if T.graph.get("family") not in ("chimera", "pegasus", "zephyr"):
        raise NotImplementedError(
            "This strategy is currently only implemented for Chimera, Pegasus"
            " and Zephyr graphs constructed by dwave_networkx`.")

    # Bin vertices of S and T into a grid graph G
    G = _intersection_binning(S_layout, T)

    placement = {}
    for _, data in G.nodes(data=True):
        for v in data["variables"]:
            placement[v] = data["qubits"]

    return placement


def _intersection_binning(S_layout, T):
    """Map the vertices of S to the "intersection graph" of T. This modifies the 
    grid graph G by assigning vertices from S and T to vertices of G.

    Args:
        S_layout (:class:`.Layout`):
            A layout for S; i.e. a map from S to R^d.
        T (networkx.Graph):
            The target graph to embed S in.

    Returns:
        networkx.Graph: A grid graph. Each vertex of G contains data attributes 
        "variables" and "qubits", that is, respectively vertices of S and T 
        assigned to that vertex.
    
    """
    # Scale the layout so that for each unit-cell edge, we have an integer point.
    m, n, t = T.graph["rows"], T.graph["columns"], T.graph["tile"]
    if T.graph['family'] == 'zephyr':
        m = 2*m+1
        n = 2*n+1
        t = 2*t

    # --- Make the "intersection graph" of the dnx_graph
    # Grid points correspond to intersection rows and columns of the dnx_graph
    G = nx.grid_2d_graph(t*n, t*m)

    # Determine the scale for putting things in the positive quadrant
    scale = (t*min(n, m) - 1)/2

    # Get the row, column mappings for the dnx graph
    lattice_mapping = _lookup_intersection_coordinates(T)

    # Less efficient, but more readable to initialize all at once
    for v in G:
        G.nodes[v]["qubits"] = set()
        G.nodes[v]["variables"] = set()

    # Add qubits (vertices of T) to grid points
    for int_point, Q in lattice_mapping.items():
        G.nodes[int_point]["qubits"] |= Q

    # --- Map the S_layout to the grid
    # "Zoom in" on layout_S so that the integer points are better represented
    zoom_scale = S_layout.scale*t
    if zoom_scale < scale:
        S_layout.scale = zoom_scale
    else:
        S_layout.scale = scale

    # Center to the positive orthant
    S_layout.center = 2*(scale, )

    # Add "variables" (vertices from S) to grid points too
    for v, pos in S_layout.items():
        grid_point = tuple(int(x) for x in np.round(pos))
        G.nodes[grid_point]["variables"].add(v)

    return G


def _lookup_intersection_coordinates(G):
    """For a dwave_networkx graph G, this returns a dictionary mapping the 
    lattice points to sets of vertices of G. 
    
    For Chimera, Pegasus and Zephyr, each lattice point corresponds to the 2
    qubits intersecting at that point.
    """
    graph_data = G.graph
    family = graph_data.get("family")
    data_key = None
    intersection_points = defaultdict(set)
    if family == "chimera":
        shore = graph_data.get("tile")
        collect_intersection_points = _chimera_all_intersection_points
        if graph_data["labels"] == "coordinate":
            def get_coords(v):
                return v
        elif graph_data["data"]:
            data_key = "chimera_index"
        else:
            coords = dnx.chimera_coordinates(
                graph_data['rows'],
                n=graph_data['columns'],
                t=shore
            )
            get_coords = coords.linear_to_chimera
    elif family == "pegasus":
        shore = [graph_data['vertical_offsets'],
                 graph_data['horizontal_offsets']]
        collect_intersection_points = _pegasus_all_intersection_points
        if graph_data["labels"] == "coordinate":
            def get_coords(v):
                return v
        elif graph_data["data"]:
            data_key = "pegasus_index"
        else:
            coords = dnx.pegasus_coordinates(graph_data['rows'])
            if graph_data['labels'] == 'int':
                get_coords = coords.linear_to_pegasus
            elif graph_data['labels'] == 'nice':
                get_coords = coords.nice_to_pegasus
    elif family == "zephyr":
        shore = graph_data.get("tile")
        collect_intersection_points = _zephyr_all_intersection_points
        if graph_data["labels"] == "coordinate":
            def get_coords(v):
                return v
        elif graph_data["data"]:
            data_key = "zephyr_index"
        else:
            coords = dnx.zephyr_coordinates(
                graph_data['rows'],
                t=shore
            )
            get_coords = coords.linear_to_zephyr

    if data_key is None:
        for v in G:
            collect_intersection_points(
                intersection_points, shore, v, *get_coords(v))
    else:
        for v, d in G.nodes(data=True):
            collect_intersection_points(
                intersection_points, shore, v, *d[data_key])

    return intersection_points

def _chimera_all_intersection_points(intersection_points, t, v, i, j, u, k):
    """Given a coordinate vertex, v = (i, j, u, k), of a Chimera with tile, t, 
    get all intersection points it is in.
    """
    # If you're a row vertex, you go in all grid points of your row intersecting columns in your unit tile
    if u == 1:
        row = i*t + k
        for kk in range(t):
            col = j*t + kk
            intersection_points[(col, row)].add(v)

    # Sameish for a column vertex.
    elif u == 0:
        col = j*t + k
        for kk in range(t):
            row = i*t + kk
            intersection_points[(col, row)].add(v)

def _pegasus_all_intersection_points(intersection_points, offsets, v, u, w, k, z):
    """Given a coordinate vertex, v = (u, w, k, z), of a Pegasus graph with offsets
    `offsets`, get all intersection points it is in.
    """
    # Each horizontal qubit spans twelve grid-points in the row 12w+k
    if u == 1:
        row = 12*w + k
        col_0 = 12*z + offsets[u][k]
        for kk in range(12):
            intersection_points[(col_0 + kk, row)].add(v)

    # Sameish for a column vertex.
    elif u == 0:
        col = 12*w + k
        row_0 = 12*z + offsets[u][k]
        for kk in range(12):
            intersection_points[(col, row_0 + kk)].add(v)

def _zephyr_all_intersection_points(intersection_points, t, v, u, w, k, j, z):
    """Given a coordinate vertex, v = (u, w, k, j, z), of a Zephyr graph with tile
    `t`, get all intersection points it is in.
    """
    if u == 0:
        row = 2*(t*w + k) + j
        col_0 = (4*z + 2*j)*t
        for kk in range(4*t):
            intersection_points[(col_0 + kk, row)].add(v)

    elif u == 1:
        col = 2*(t*w + k) + j
        row_0 = (4*z + 2*j)*t
        for kk in range(4*t):
            intersection_points[(col, row_0 + kk)].add(v)

def closest(S_layout, T_layout, subset_size=(1, 1), num_neighbors=1, **kwargs):
    """Maps vertices of S to the closest vertices of T as given by ``S_layout`` 
    and ``T_layout``. i.e. For each vertex u in ``S_layout`` and each vertex v 
    in ``T_layout``, map u to the v with minimum Euclidean distance 
    :math:`(||u - v||_2)`.

    By default, :func:`closest` is used to compute the placement of an embedding
    when :func:`minorminer.layout.find_embedding` is called.

    Args:
        S_layout (:class:`.Layout`):
            A layout for S; i.e. a map from S to :math:`R^d`.

        T_layout (:class:`.Layout`):
            A layout for T; i.e. a map from T to :math:`R^d`.

        subset_size (tuple, optional, default=(1, 1)):
            A lower (subset_size[0]) and upper (subset_size[1]) bound on the size 
            of subsets of T that will be considered when mapping vertices of S.
        
        num_neighbors (int, optional, default=1):
            The number of closest neighbors to query from the KDTree--the 
            neighbor with minimum overlap is chosen. Increasing this reduces 
            overlap, but increases runtime.

    Returns:
        dict: A mapping from vertices of S (keys) to subsets of vertices of T (values).

    Examples:
        This example creates a :class:`.Placement` object that stores a mapping 
        computed with :func:`.closest`, in which the nodes from a source 
        hexagonal lattice graph are mapped to a target Chimera graph.

        >>> import networkx as nx
        >>> import dwave_networkx as dnx
        >>> import minorminer.layout as mml
        ...
        >>> G = nx.hexagonal_lattice_graph(2,2)
        >>> G_layout = mml.Layout(G, mml.p_norm)
        >>> C = dnx.chimera_graph(2,2)
        >>> C_layout = mml.Layout(C, mml.dnx_layout)
        >>> placement = mml.Placement(G_layout, C_layout, placement=mml.closest)

    """
    # Extract the target graph
    T = T_layout.G

    # A new layout for subsets of T.
    T_subgraph_layout = {}

    # Get connected subgraphs to consider mapping to
    T_subgraphs = _get_connected_subgraphs(T, subset_size[1])

    # Calculate the barycenter (centroid) of each subset
    for k in range(subset_size[0], subset_size[1]+1):
        if k == 1:
            for subgraph in T_subgraphs[k]:
                v, = subgraph  # Unpack the subgraph of size 1
                T_subgraph_layout[subgraph] = T_layout[v]
        else:
            for subgraph in T_subgraphs[k]:
                T_subgraph_layout[subgraph] = np.mean(
                    np.array([T_layout[v] for v in subgraph]), axis=0)

    # Use scipy's KDTree to solve the nearest neighbor problem.
    # This requires a few lookup tables
    T_subset_lookup = {tuple(p): V for V, p in T_subgraph_layout.items()}
    layout_points = [tuple(p) for p in T_subgraph_layout.values()]
    overlap_counter = Counter()

    try:
        tree = spatial.KDTree(layout_points)  # This fails for the empty graph
    except ValueError:
        pass

    placement = {}
    for u, u_pos in S_layout.items():
        distances, v_indices = tree.query([u_pos], num_neighbors)

        # KDTree.query either returns a (num_neighbors, ) shaped arrays if num_neighbors == 1
        # or (1, num_neighbors) shaped arrays if num_neighbors != 1
        if num_neighbors != 1:
            v_indices = v_indices[0]
            distances = distances[0]

        placement[u] = _minimize_overlap(
            distances, v_indices, T_subset_lookup, layout_points, overlap_counter)

    return placement


def _get_connected_subgraphs(G, k, single_set=False):
    """Finds all connected subgraphs S of G within a given subset_size.

    Args:
        G (networkx.Graph):
            The graph you want to find all connected subgraphs of.
        k (int):
            An upper bound of the size of connected subgraphs to find.

    Returns:
        dict: connected_subgraphs, the dictionary is keyed by size of subgraph 
        and each value is a set containing frozensets of vertices that comprise 
        the connected subgraphs.
        {
            1: { {v_1}, {v_2}, ... },
            2: { {v_1, v_2}, {v_1, v_3}, ... },
            ...,
            k: { {v_1, v_2, ..., v_m}, ... }
        }

    """
    connected_subgraphs = defaultdict(set)
    connected_subgraphs[1] = {frozenset((v,)) for v in G}

    for i in range(1, min(k, len(G))):
        # Iterate over the previous set of connected subgraphs.
        for X in connected_subgraphs[i]:
            # For each vertex in the set, iterate over its neighbors.
            for v in X:
                for u in G.neighbors(v):
                    connected_subgraphs[i + 1].add(X.union({u}))

    return connected_subgraphs


def _minimize_overlap(distances, v_indices, T_subset_lookup, layout_points, overlap_counter):
    """A greedy penalty-type model for choosing nonoverlapping chains."""
    subsets = {}
    for i, d in zip(v_indices, distances):
        subset = T_subset_lookup[layout_points[i]]
        subsets[subset] = d + sum(10**overlap_counter[v] for v in subset)

    cheapest_subset = min(subsets, key=subsets.get)
    overlap_counter.update(cheapest_subset)
    return cheapest_subset


class Placement(abc.MutableMapping):
    """Class that stores (or computes) a mapping of source nodes to collections 
    of target nodes without any constraints. In mathematical terms, map V(S) to 
    :math:`2^{V(T)}`.

    Args:
        S_layout (:class:`.Layout`):
            A layout for S; i.e. a map from S to :math:`R^d`.

        T_layout (:class:`.Layout`):
            A layout for T; i.e. a map from T to :math:`R^d`.

        placement (dict/function, optional, default=None):
            If a dict, this specifies a pre-computed placement for S in T.
            
            If a function, the function is called on ``S_layout`` and ``T_layout``, 
            ``placement(S_layout, T_layout)``, and should return a dict representing 
            a placement of S in T. 
            
            If None, a random placement of S in T is selected.

        scale_ratio (float, optional, default=None):
            If None, ``S_layout`` is not scaled. Otherwise, ``S_layout`` is scaled 
            to ``scale_ratio*T_layout.scale``.

        **kwargs (dict):
            Keyword arguments are passed to ``placement`` if it is a function.
    
    """
    def __init__(
        self,
        S_layout,
        T_layout,
        placement=None,
        scale_ratio=None,
        **kwargs
    ):
        self.S_layout = _parse_layout(S_layout)
        self.T_layout = _parse_layout(T_layout)

        # Layout dimensions should match
        if self.S_layout.dim != self.T_layout.dim:
            raise ValueError(
                "S_layout has dimension {} but T_layout has dimension {}. These must match.".format(
                    self.S_layout.dim, self.T_layout.dim)
            )

        # Scale S if S_layout is bigger than T_layout
        if self.S_layout.scale > self.T_layout.scale:
            self.S_layout.scale = self.T_layout.scale
        # Or scale S to the user specified scale
        elif scale_ratio:
            self.S_layout.scale = scale_ratio*self.T_layout.scale

        if placement is None:
            self.placement = closest(self.S_layout, self.T_layout)
        elif callable(placement):
            self.placement = placement(self.S_layout, self.T_layout, **kwargs)
        else:
            self.placement = placement

    # The class should behave like a dictionary
    def __iter__(self):
        """Iterate through the keys of the dictionary placement."""
        yield from self.placement

    def __getitem__(self, key):
        """Get the placement value at the key vertex."""
        return self.placement[key]

    def __setitem__(self, key, value):
        """Set the placement value at the key vertex."""
        self.placement[key] = value

    def __delitem__(self, key):
        """Delete the placement value at the key vertex."""
        del self.placement[key]

    def __repr__(self):
        """Use the placement's dictionary representation."""
        return repr(self.placement)

    def __len__(self):
        """The length of a placement is the length of the placement dictionary."""
        return len(self.placement)


def _parse_layout(G_layout):
    """Ensures a Layout object was passed in and makes a copy to save in the 
    Placement object.
    """
    if isinstance(G_layout, layout.Layout):
        return layout.Layout(G_layout.G, G_layout.layout)

    else:
        raise TypeError(
            "If you want to pass in a precomputed layout mapping, please create a Layout object; Layout(G, layout).")



# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import time

import networkx as nx
import minorminer as mm

from .layout import Layout, dnx_layout, p_norm
from .placement import Placement, closest, intersection


def find_embedding(
    S,
    T,
    layout=(p_norm, None),
    placement=closest,
    mm_hint_type="initial_chains",
    return_layouts=False,
    **kwargs
):
    """Tries to embed S in T by computing layout-aware chains and passing 
    them to :func:`minorminer.find_embedding`. Chains are passed as either 
    `initial_chains` or `suspend_chains` (see documentation for 
    :func:`minorminer.find_embedding` to learn more).

    Args:
        S (NetworkX Graph/edges data structure (dict, list, ...)):
            The source graph being embedded or a NetworkX supported data structure
            for edges (see :func:`nx.convert.to_networkx_graph` for details).

        T (NetworkX Graph/edges data structure (dict, list, ...)):
            The target graph being embedded into or a NetworkX supported data 
            structure for edges (see :func:`nx.convert.to_networkx_graph` for 
            details).

        layout (function/(function/dict/Layout, function/dict/Layout), optional):
            A function to compute the :class:`.Layout` for both S and T, or a 
            2-tuple that either consists of a pair of functions or pre-computed 
            layouts (in the form of :class:`.Layout` or dicts). The first entry 
            in the 2-tuple applies to S while the second applies to T.

            Note:
                If ``layout`` is a single function and T is a dnx_graph, then the 
                function passed in is only applied to S and the dnx_layout is 
                applied to T. To run a layout function explicitly on T, pass 
                it in as a 2-tuple; i.e. (p_norm, p_norm).

        placement (function/dict, optional, default=minorminer.placement.closest):
            A function that uses the layouts of S and T to map the vertices 
            of S to subsets of vertices of T (:class:`.Placement`), or a dict 
            that contains the precomputed mapping/:class:`.Placement`.
            
            By default, :func:`~placement.closest` is called to compute placement.

        mm_hint_type (str, optional, default="initial_chains"):
            This is the hint type passed to :func:`minorminer.find_embedding`. 
            Supported types are "initial_chains" and "suspend_chains". See
            :func:`minorminer.find_embedding` for more information.
        
        return_layouts (bool, optional, default=False):
            If True, layout objects of S and T are also returned.

        **kwargs (dict):
            Keyword arguments passed to :class:`.Layout`, :class:`.Placement` 
            or :func:`minorminer.find_embedding`.

    Returns:
        dict: An embedding of vertices of S (keys) to chains in T (values). This 
        embedding is dependent on the kwargs being passed in. If ``return_layouts``
        is True, a 2-tuple is returned in which the first element is the embedding 
        dict and the second element is another 2-tuple containing the source and 
        target :class:`.Layout` objects.

    """
    start = time.perf_counter()

    # Parse kwargs
    layout_kwargs, placement_kwargs = _parse_kwargs(kwargs)

    # Compute the layouts
    S_layout, T_layout = _parse_layout_parameter(S, T, layout, layout_kwargs)

    # Compute the placement (i.e. chains)
    S_T_placement = Placement(
        S_layout, T_layout, placement, **placement_kwargs)

    end = time.perf_counter()
    timeout = kwargs.get("timeout")
    if timeout is not None:
        time_remaining = timeout - (end - start)
        if time_remaining <= 0:
            raise TimeoutError(
                "Layout and placement took {}s, you set timeout at {}.".format(end-start, timeout))
        kwargs["timeout"] = time_remaining

    # Run minorminer.find_embedding
    if mm_hint_type == "initial_chains":
        output = mm.find_embedding(
            S, T, initial_chains=S_T_placement, **kwargs)
    elif mm_hint_type == "suspend_chains":
        output = mm.find_embedding(S, T, suspend_chains={
            v: [C] for v, C in S_T_placement.items()}, **kwargs)
    else:
        raise ValueError(
            "Only initial_chains and suspend_chains are supported minorminer hint types.")

    if return_layouts:
        return output, (S_layout, T_layout)
    return output


def _parse_kwargs(kwargs):
    """Extract kwargs for layout and placement functions. Leave the remaining 
    ones for :func:`minorminer.find_embedding`.
    """
    layout_kwargs = {}
    # For the layout object
    if "dim" in kwargs:
        layout_kwargs["dim"] = kwargs.pop("dim")
    if "center" in kwargs:
        layout_kwargs["center"] = kwargs.pop("center")
    if "scale" in kwargs:
        layout_kwargs["scale"] = kwargs.pop("scale")

    placement_kwargs = {}
    # For the placement object
    if "scale_ratio" in kwargs:
        placement_kwargs["scale_ratio"] = kwargs.pop("scale_ratio")
    # For closest strategy
    if "subset_size" in kwargs:
        placement_kwargs["subset_size"] = kwargs.pop("subset_size")
    if "num_neighbors" in kwargs:
        placement_kwargs["num_neighbors"] = kwargs.pop("num_neighbors")

    return layout_kwargs, placement_kwargs


def _parse_layout_parameter(S, T, layout, layout_kwargs):
    """Determine what combination of iterable, dict, and function the layout 
    parameter is.
    """
    if nx.utils.iterable(layout):
        try:
            s_layout, t_layout = layout
        except ValueError:
            raise ValueError(
                "layout is expected to be a function or a length-2 iterable")
    else:
        s_layout = t_layout = layout

    # Get a Layout object for S
    if isinstance(s_layout, Layout):
        S_layout = s_layout
    else:
        # Assumes s_layout a callable or implements a mapping interface
        S_layout = Layout(S, layout=s_layout, **layout_kwargs)

    # Get the Layout object for T
    if isinstance(t_layout, Layout):
        T_layout = t_layout
    else:
        # Use the dnx_layout if possible
        if T.graph.get("family") in ("chimera", "pegasus", "zephyr"):
            T_layout = Layout(T, layout=dnx_layout, **layout_kwargs)
        # Assumes t_layout a callable or implements a mapping interface
        else:
            T_layout = Layout(T, layout=t_layout, **layout_kwargs)

    return S_layout, T_layout



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================


class EmbeddingError(Exception):
    """Base class for all embedding exceptions."""


class MissingChainError(EmbeddingError):
    """Raised if a node in the source graph has no associated chain.

    Args:
        snode: The source node with no associated chain.

    """
    def __init__(self, snode):
        msg = "chain for {} is empty or not contained in this embedding"
        super(MissingChainError, self).__init__(msg.format(snode))
        self.source_node = snode


class ChainOverlapError(EmbeddingError):
    """Raised if two source nodes have an overlapping chain.

    Args:
        tnode: Location where the chains overlap.
        snode0: First source node with overlapping chain.
        snode1: Second source node with overlapping chain.

    """
    def __init__(self, tnode, snode0, snode1):
        msg = "overlapped chains at target node {}: source nodes are {} and {}"
        super(ChainOverlapError, self).__init__(msg.format(tnode, snode0, snode1))
        self.target_node = tnode
        self.source_nodes = (snode0, snode1)


class DisconnectedChainError(EmbeddingError):
    """Raised if a chain is not connected in the target graph.

    Args:
        snode: The source node associated with the broken chain.

    """
    def __init__(self, snode):
        msg = "chain for {} is not connected"
        super(DisconnectedChainError, self).__init__(msg.format(snode))
        self.source_node = snode


class InvalidNodeError(EmbeddingError):
    """Raised if a chain contains a node not in the target graph.

    Args:
        snode: The source node associated with the chain.
        tnode: The node in the chain not in the target graph.

    """
    def __init__(self, snode, tnode):
        msg = "chain for {} contains a node label {} not contained in the target graph"
        super(InvalidNodeError, self).__init__(msg.format(snode, tnode))
        self.source_node = snode
        self.target_node = tnode


class MissingEdgeError(EmbeddingError):
    """Raised when two source nodes sharing an edge to not have a corresponding edge between their chains.

    Args:
        snode0: First source node.
        snode1: Second source node.

    """
    def __init__(self, snode0, snode1):
        msg = "source edge ({}, {}) is not represented by any target edge"
        super(MissingEdgeError, self).__init__(msg.format(snode0, snode1))
        self.source_nodes = (snode0, snode1)



# Copyright 2016 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================
"""

**This module is used as a private source file. It is subject to change and not
recommended for use at this time.**

This file implements a polynomial-time algorithm to find a maximum-sized
native clique embedding in an induced Chimera subgraph.  It also provides
functionality to find maximum-sized native embeddings of complete bipartite
graphs (bicliques).  Additionally, we wrap the polynomial-time algorithm to
compute maximum-sized native (bi)clique embeddings for non-induced subgraphs
(that is, for processors which have both qubit and coupler failures).

The :class:`eden_processor` class provides the implementation of the polynomial-time
algorithm described in [BKR]_.  That algorithm relies, in part, on building a
cache of the number of "horizontal wires" and "vertical wires" across a
processor (that is, paths of horizontally- and vertically- aligned qubits) in
all of the rectangular sub-regions of the processor.  As a side-effect, we can
use the same cache to look at bicliques; so that functionality is also built
into the :class:`eden_processor` class.

The :class:`eden_processor` class is written under the assumption that if a pair of
qubits exists in the processor and the Chimera structure implies that those
qubits are connected by a coupler, then that coupler exists (that is, we
assume that the graph is an induced Chimera subgraph).  That assumption
is not true for all processors, so we need to accomodate this somehow.

The :class:`processor` class is suitable for end-users.  It can be used to either
produce optimal native embeddings, or approximately-optimal native embeddings.


.. [BKR] Kelly Boothby, Andrew D. King, Aidan Roy.  Fast clique minor
    generation in Chimera qubit connectivity graphs. Quantum Information
    Processing, (2015).  http://arxiv.org/abs/1507.04774

"""

#TODO fix this or delete the entire file
__bugged_graph = [(12, 9), (12, 4), (12, 8), (12, 11), (12, 10), (15, 9), (15, 7), (15, 8), (15, 11), (15, 10), (30, 27), (30, 24), (30, 26), (30, 25), (2, 5), (2, 4), (2, 7), (2, 6), (16, 22), (16, 21), (16, 20), (16, 23), (5, 1), (5, 3), (5, 0), (19, 22), (19, 21), (19, 23), (22, 18), (22, 17), (9, 14), (9, 13), (27, 29), (27, 31), (27, 28), (24, 29), (24, 31), (24, 28), (29, 21), (29, 26), (29, 25), (14, 8), (14, 11), (14, 6), (1, 4), (1, 7), (1, 6), (4, 3), (18, 21), (18, 20), (18, 23), (7, 3), (7, 0), (21, 17), (8, 13), (26, 31), (26, 28), (13, 10), (31, 25), (3, 6), (28, 25), (6, 0), (20, 17), (23, 17)]

from random import shuffle, randint, choice, sample
from collections import defaultdict
from itertools import product
import warnings

warnings.warn("minorminer.utils.polynomialembedder is deprecated. Use minorminer.busclique intead.", DeprecationWarning)

__author__ = "Kelly Boothby"

__all__ = ['eden_processor', 'processor', 'random_processor']


def _accumulate_random(count, found, oldthing, newthing):
    """This performs on-line random selection.

    We have a stream of objects

        o_1,c_1; o_2,c_2; ...

    where there are c_i equivalent objects like o_1.  We'd like to pick
    a random object o uniformly at random from the list

        [o_1]*c_1 + [o_2]*c_2 + ...

    (actually, this algorithm allows arbitrary positive weights, not
    necessarily integers)  without spending the time&space to actually
    create that list.  Luckily, the following works:

        thing = None
        c_tot
        for o_n, c_n in things:
            c_tot += c_n
            if randint(1,c_tot) <= c_n:
                thing = o_n

    This function is written in an accumulator format, so it can be
    used one call at a time:

    EXAMPLE:
        > thing = None
        > count = 0
        > for i in range(10):
        >     c = 10-i
        >     count, thing = accumulate_random(count,c,thing,i)


    INPUTS:
        count: integer, sum of weights found before newthing
        found: integer, weight for newthing
        oldthing: previously selected object (will never be selected
                  if count == 0)
        newthing: incoming object

    OUTPUT:
        (newcount, pick): newcount is count+found, pick is the newly
                          selected object.
    """
    if randint(1, count + found) <= found:
        return count + found, newthing
    else:
        return count + found, oldthing


class eden_processor(object):
    """Class to hold a processor, and embed cliques in said processor.

    Syntactic sugar: (assume proc is a :class:`eden_processor` instance)

    >>> q in proc # q is a working qubit
    True
    >>> proc[q] # returns the neighbors of q connected by a working coupler
    set([...])

    **CAVEAT**: the embedding utilities herein assume that the only failed
    couplers connected to working qubits are between two cells.  Any
    pair of working qubits on opposite sides of the same unit cell are
    assumed to be connected by working couplers.
    """

    def __init__(self, edgelist, M, N, L, random_bundles=False):
        """Constructs an :class:`eden_processor` instance

        Args:
            edgelist (list): a list of tuples (p,q) where each of p and q is
                a qubit address in chimera coordinates.  These are taken to be
                the working couplers.
            M (int): first chimera parameter, :math:`C_{M,N,L}`
            N (int): as above
            L (int): as above
            random_bundles (bool): whether or not to shuffle horizontal and
                vertical line bundles when assembling ells.
        """
        couplers = defaultdict(set)
        for p, q in edgelist:
            couplers[p].add(q)
            couplers[q].add(p)
        self._couplers = couplers
        self.M = M
        self.N = N
        self.L = L
        self._compute_vline_scores()
        self._compute_hline_scores()
        self.random_bundles = random_bundles
        self._biclique_size = {}
        self._biclique_size_computed = False

    def __getitem__(self, q):
        """Return the qubits coupled to `q`.
        """
        return self._couplers[q]

    def __contains__(self, q):
        """Check if the qubit `q` is contained in this processor.
        """
        return q in self._couplers

    def __iter__(self):
        """Iterate over the qubits in this processor.
        """
        for q in self._couplers:
            yield q

    def vline_score(self, x, ymin, ymax):
        """Returns the number of unbroken paths of qubits

        >>> [(x,y,1,k) for y in range(ymin,ymax+1)]

        for :math:`k = 0,1,\cdots,L-1`.  This is precomputed for speed.
        """
        return self._vline_score[x, ymin, ymax]

    def hline_score(self, y, xmin, xmax):
        """Returns the number of unbroken paths of qubits

        >>> [(x,y,0,k) for x in range(xmin,xmax+1)]

        for :math:`k = 0,1,\cdots,L-1`.  This is precomputed for speed.
        """
        return self._hline_score[y, xmin, xmax]

    def _compute_vline_scores(self):
        """Does the hard work to prepare ``vline_score``.
        """
        M, N, L = self.M, self.N, self.L
        vline_score = {}
        for x in range(M):
            laststart = [0 if (x, 0, 1, k) in self else None for k in range(L)]
            for y in range(N):
                block = [0] * (y + 1)
                for k in range(L):
                    if (x, y, 1, k) not in self:
                        laststart[k] = None
                    elif laststart[k] is None:
                        laststart[k] = y
                        block[y] += 1
                    elif y and (x, y, 1, k) not in self[x, y - 1, 1, k]:
                        laststart[k] = y
                    else:
                        for y1 in range(laststart[k], y + 1):
                            block[y1] += 1
                for y1 in range(y + 1):
                    vline_score[x, y1, y] = block[y1]
        self._vline_score = vline_score

    def _compute_hline_scores(self):
        """Does the hard work to prepare ``hline_score``.
        """
        M, N, L = self.M, self.N, self.L
        hline_score = {}
        for y in range(N):
            laststart = [0 if (0, y, 0, k) in self else None for k in range(L)]
            for x in range(M):
                block = [0] * (x + 1)
                for k in range(L):
                    if (x, y, 0, k) not in self:
                        laststart[k] = None
                    elif laststart[k] is None:
                        laststart[k] = x
                        block[x] += 1
                    elif x and (x, y, 0, k) not in self[x - 1, y, 0, k]:
                        laststart[k] = x
                    else:
                        for x1 in range(laststart[k], x + 1):
                            block[x1] += 1
                for x1 in range(x + 1):
                    hline_score[y, x1, x] = block[x1]
        self._hline_score = hline_score

    def _compute_biclique_sizes(self, recompute=False):
        """Calls ``self.biclique_size(...)`` for every rectangle contained in this
        processor, to fill the biclique size cache.

        INPUTS:
            recompute: if ``True``, then we dump the existing cache and compute
                all biclique sizes from scratch.  (default: ``False``)
        """
        if recompute or not self._biclique_size_computed:
            self._biclique_size = {}
            self._biclique_size_to_length = defaultdict(dict)
            self._biclique_length_to_size = defaultdict(dict)
        else:
            return
        M, N = self.M, self.N
        for xmax in range(M):
            for xmin in range(xmax + 1):
                for ymax in range(N):
                    for ymin in range(ymax + 1):
                        ab = self.biclique_size(xmin, xmax, ymin, ymax)
                        wh = xmax - xmin + 1, ymax - ymin + 1
                        self._biclique_size_to_length[ab][
                            wh] = (xmin, xmax, ymin, ymax)
                        self._biclique_length_to_size[wh][
                            ab] = (xmin, xmax, ymin, ymax)

        self._biclique_size_computed = True

    def biclique_size(self, xmin, xmax, ymin, ymax):
        """Returns the size parameters ``(m,n)`` of the complete bipartite graph
        :math:`K_{m,n}` comprised of ``m`` unbroken chains of horizontally-aligned qubits
        and ``n`` unbroken chains of vertically-aligned qubits (known as line
        bundles)

        INPUTS:
            xmin,xmax,ymin,ymax: integers defining the bounds of a rectangle
            where we look for unbroken chains.  These ranges include both
            endpoints.

        OUTPUTS:
            m,n: integers corresponding to the number of horizontal and
                vertical line bundles contained in this rectangle.
        """
        try:
            return self._biclique_size[xmin, xmax, ymin, ymax]
        except KeyError:
            hscore = self.hline_score(ymin, xmin, xmax)
            vscore = self.vline_score(xmin, ymin, ymax)
            if ymin < ymax:
                hscore += self.biclique_size(xmin, xmax, ymin + 1, ymax)[0]
            if xmin < xmax:
                vscore += self.biclique_size(xmin + 1, xmax, ymin, ymax)[1]
            self._biclique_size[xmin, xmax, ymin, ymax] = hscore, vscore
            return hscore, vscore

    def biclique(self, xmin, xmax, ymin, ymax):
        """Compute a maximum-sized complete bipartite graph contained in the
        rectangle defined by ``xmin, xmax, ymin, ymax`` where each chain of
        qubits is either a vertical line or a horizontal line.

        INPUTS:
            xmin,xmax,ymin,ymax: integers defining the bounds of a rectangle
            where we look for unbroken chains.  These ranges include both
            endpoints.

        OUTPUT:
            (A_side, B_side): a tuple of two lists containing lists of qubits.
                the lists found in ``A_side`` and ``B_side`` are chains of qubits.
                These lists of qubits are arranged so that

                >>> [zip(chain,chain[1:]) for chain in A_side]

                and

                >>> [zip(chain,chain[1:]) for chain in B_side]

                are lists of valid couplers.
        """

        Aside = sum((self.maximum_hline_bundle(y, xmin, xmax)
                     for y in range(ymin, ymax + 1)), [])
        Bside = sum((self.maximum_vline_bundle(x, ymin, ymax)
                     for x in range(xmin, xmax + 1)), [])
        return Aside, Bside


    def _contains_line(self, line):
        """Test if a chain of qubits is completely contained in ``self``.  In
        particular, test if all qubits are present and the couplers
        connecting those qubits are also connected.

        NOTE: this function assumes that ``line`` is a list or tuple of
        qubits which satisfies the precondition that ``(line[i],line[i+1])``
        is supposed to be a coupler for all ``i``.

        INPUTS:
            line: a list of qubits satisfying the above precondition

        OUTPUT:
            boolean
        """
        return all(v in self for v in line) and all(u in self[v] for u, v in zip(line, line[1::]))

    def maximum_vline_bundle(self, x0, y0, y1):
        """Compute a maximum set of vertical lines in the unit cells ``(x0,y)``
        for :math:`y0 \leq y \leq y1`.

        INPUTS:
            y0,x0,x1: int

        OUTPUT:
            list of lists of qubits
        """

        y_range = range(y1, y0 - 1, -1) if y0 < y1 else range(y1, y0 + 1)
        vlines = [[(x0, y, 1, k) for y in y_range] for k in range(self.L)]
        return list(filter(self._contains_line, vlines))

    def maximum_hline_bundle(self, y0, x0, x1):
        """Compute a maximum set of horizontal lines in the unit cells ``(x,y0)``
        for :math:`x0 \leq x \leq x1`.

        INPUTS:
            y0,x0,x1: int

        OUTPUT:
            list of lists of qubits
        """
        x_range = range(x0, x1 + 1) if x0 < x1 else range(x0, x1 - 1, -1)
        hlines = [[(x, y0, 0, k) for x in x_range] for k in range(self.L)]
        return list(filter(self._contains_line, hlines))

    def maximum_ell_bundle(self, ell):
        """Return a maximum ell bundle in the rectangle bounded by

            :math:`\{x0,x1\} \\times \{y0,y1\}`

        with vertical component

            :math:`(x0,y0) ... (x0,y1) = {x0} \\times \{y0,...,y1\}`

        and horizontal component

            :math:`(x0,y0) ... (x1,y0) = \{x0,...,x1\} \\times \{y0\}`.

        Note that we don't require :math:`x0 \leq x1` or :math:`y0 \leq y1`.  We go
        through some shenanigans so that the qubits we return
        are all in a path.  A nice side-effect of this is that

            >>> chains = maximum_ell_bundle(...)
            >>> edges = [zip(path,path[:-1]) for path in chains]

        where ``edges`` will be a list of lists of chain edges.

        INPUTS::
            ell: a tuple of 4 integers defining the ell, ``(x0, x1, y0, y1)``

        OUTPUT::
            chains: list of lists of qubits

        Note: this function only to be called to construct a
        native clique embedding *after* the block embedding has
        been constructed.  Using this to evaluate the goodness
        of an ell block will be slow.
        """
        (x0, x1, y0, y1) = ell
        hlines = self.maximum_hline_bundle(y0, x0, x1)
        vlines = self.maximum_vline_bundle(x0, y0, y1)

        if self.random_bundles:
            shuffle(hlines)
            shuffle(vlines)

        return [v + h for h, v in zip(hlines, vlines)]

    def _combine_clique_scores(self, rscore, hbar, vbar):
        """Computes the score of a partial native clique embedding given the score
        attained by the already-placed ells, together with the ell block
        defined by ``hbar = (y0,xmin,xmax)``, and  ``vbar = (x0,ymin,ymax)``.

        In the plain :class:`eden_processor` class, this is simply the number of ells
        contained in the partial native clique after adding the new ells.
        """
        (y0, xmin, xmax) = hbar
        (x0, ymin, ymax) = vbar
        if rscore is None:
            rscore = 0
        hscore = self.hline_score(y0, xmin, xmax)
        vscore = self.vline_score(x0, ymin, ymax)
        if vscore < hscore:
            score = rscore + vscore
        else:
            score = rscore + hscore
        return score

    def maxCliqueWithRectangle(self, R, maxCWR):
        """This does the dirty work for :func:`nativeCliqueEmbed`.  Not meant to be
        understood or called on its own.  Guaranteed to maintain the inductive
        hypothesis that ``maxCWR`` is optimal.  We put in the tiniest amount of
        effort to return a uniform random choice of a maximum-sized native
        clique embedding.

        INPUTS:
            R (tuple): the rectangle specified as a tuple ``(xmin,xmax,ymin,ymax)``

            maxCWR (dict): the dictionary we're building inductively which maps
            ``R -> (score, ell, r, num)`` where score is the size of the best
            clique with working rectangle ``R``, and ``ell`` and ``r`` indicate how to
            reconstruct that clique: add the maximum line bundle in ``ell``, and
            recursively examine the working rectangle ``r``.  The parameter ``num`` is the number
            of cliques with ``R`` as a working rectangle and size equal to ``score``.

        OUTPUT:
            score (int): the score for the returned clique (just ``len(clique)``
            in the class :class:`eden_processor`; may differ in subclasses)

            best (tuple): a tuple ``(score, ell, parent, num)`` to be stored in
            ``maxCWR[R]``.

            * score: as above
            * ell: ``(x0,x1,y0,y1)`` defines the unit cells in an ell-shaped
                region: ``(x0,y1),...,(x0,y0),...,(x1,y0)``
            * parent: the rectangle ``R1`` for which the clique generated
                recursively by looking up ``maxCWR[R1]`` as described above.
            * num: the number of partial native clique embeddings with ``R``
                as a working rectangle and a score of ``score``.

        """

        (xmin, xmax, ymin, ymax) = R
        best = nothing = 0, None, None, 1
        bestscore = None
        count = 0
        N = self.N

        Xlist = (xmin, xmax, xmin + 1, xmax), (xmax, xmin, xmin, xmax - 1)
        Ylist = (ymin - 1, ymax, ymin - 1,
                 ymax), (ymax + 1, ymin, ymin, ymax + 1)
        XY = [(X, Y) for X in Xlist for Y in Ylist if 0 <= Y[2] <= Y[3] < N]

        bests = []
        for X, Y in XY:
            x0, x1, nxmin, nxmax = X
            y0, y1, nymin, nymax = Y
            r = nxmin, nxmax, nymin, nymax
            try:
                rscore, rell, rparent, nr = maxCWR[r]
            except:
                rscore, nr = None, 1
            score = self._combine_clique_scores(
                rscore, (y0, xmin, xmax), (x0, nymin, nymax))
            if bestscore is None or score > bestscore:
                bestscore = score
                count = 0
            if score == bestscore:
                count, best = _accumulate_random(
                    count, nr, best, (score, (x0, x1, y0, y1), r, nr))

        return bestscore, best

    def nativeCliqueEmbed(self, width):
        """Compute a maximum-sized native clique embedding in an induced
        subgraph of chimera with all chainlengths ``width+1``.

        INPUTS:
            width: width of the squares to search, also `chainlength`-1

        OUTPUT:
            score: the score for the returned clique (just ``len(clique)``
            in the class :class:`eden_processor`; may differ in subclasses)

            clique: a list containing lists of qubits, each associated
            to a chain.  These lists of qubits are carefully
            arranged so that

            >>> [zip(chain,chain[1:]) for chain in clique]

            is a list of valid couplers.

        """
        maxCWR = {}

        M, N = self.M, self.N
        maxscore = None
        count = 0
        key = None
        for w in range(width + 2):
            h = width - w - 2
            for ymin in range(N - h):
                ymax = ymin + h
                for xmin in range(M - w):
                    xmax = xmin + w
                    R = (xmin, xmax, ymin, ymax)
                    score, best = self.maxCliqueWithRectangle(R, maxCWR)
                    maxCWR[R] = best
                    if maxscore is None or (score is not None and maxscore < score):
                        maxscore = score
                        key = None  # this gets overwritten immediately
                        count = 0  # this gets overwritten immediately
                    if maxscore == score:
                        count, key = _accumulate_random(count, best[3], key, R)

        clique = []
        while key in maxCWR:
            score, ell, key, num = maxCWR[key]
            if ell is not None:
                meb = self.maximum_ell_bundle(ell)
                clique.extend(meb)
        return maxscore, clique

    def largestNativeClique(self, max_chain_length=None):
        """Returns the largest native clique embedding we can find on the
        processor, with the shortest chainlength possible (for that
        clique size).

        OUTPUT:
            score: the score for the returned clique (just ``len(clique)``
            in the class :class:`eden_processor`; may differ in subclasses)

            clique: a list containing lists of qubits, each associated
            to a chain.  These lists of qubits are carefully
            arranged so that

             >>> [zip(chain,chain[1:]) for chain in clique]

            is a list of valid couplers.

        CAVEAT: we assume that the only failed couplers connected to working
        qubits are between two cells.  Any pair of working qubits on opposite
        sides of the same unit cell are assumed to be connected by working
        couplers.
        """
        bigclique = []
        bestscore = None
        if max_chain_length is None:
            wmax = min(self.M, self.N)
        else:
            wmax = max_chain_length - 1

        for w in range(wmax + 1):
            score, clique = self.nativeCliqueEmbed(w)
            if bestscore is None or score > bestscore:
                bigclique = clique
                bestscore = score
        return bestscore, bigclique

    def tightestNativeClique(self, n):
        """Returns a native clique embedding with the shortest chains and
        at least ``n`` completely connected chains, or ``None`` on failure.

        INPUTS:
            n: size of clique to return

        OUTPUT:
            score: the score for the returned clique (just ``len(clique)``
            in the class :class:`eden_processor`; may differ in subclasses)

            clique: a list containing lists of qubits, each associated
            to a chain.  These lists of qubits are carefully arranged so that

            >>> [zip(chain,chain[1:]) for chain in clique]

            is a list of valid couplers.  If no ``n``-clique is available on
            the processor, this is ``None``.

        """
        for w in range(1, min(self.M, self.N) + 1):
            score, clique = self.nativeCliqueEmbed(w)
            if len(clique) >= n:
                return score, clique[:n]
        return None, []

    def largestNativeBiClique(self, chain_imbalance=0, max_chain_length=None):
        """Returns a native embedding for the complete bipartite graph :math:`K_{n,m}`
        for :math:`n \leq m`; where :math:`n` is as large as possible and :math:`m` is as large as
        possible subject to :math:`n`.  The native embedding of a complete bipartite
        graph is a set of horizontally-aligned qubits connected in lines
        together with an equal-sized set of vertically-aligned qubits
        connected in lines.

        INPUTS:
            chain_imbalance: how big of a difference to allow between the
            chain lengths on the two sides of the bipartition. If ``None``,
            then we allow an arbitrary imbalance.  (default: ``0``)

            max_chain_length: longest chain length to consider or ``None`` if chain
            lengths are allowed to be unbounded.  (default: ``None``)

        OUTPUT:
            score (tuple): the score for the returned clique (just ``(n,m)`` in the class
            :class:`eden_processor`; may differ in subclasses)

            embedding (tuple): a tuple of two lists containing lists of qubits.
            If ``embedding = (A_side, B_side)``, the lists found in ``A_side`` and
            ``B_side`` are chains of qubits.
            These lists of qubits are arranged so that

            >>> [zip(chain,chain[1:]) for chain in A_side]

            and

            >>> [zip(chain,chain[1:]) for chain in B_side]

            are lists of valid couplers.

        """
        self._compute_biclique_sizes()
        Len2Siz = self._biclique_length_to_size
        Siz2Len = self._biclique_size_to_length
        overkill = self.M + self.N
        if max_chain_length is None:
            max_chain_length = overkill
        if chain_imbalance is None:
            chain_imbalance = overkill

        def acceptable_chains(t):
            a, b = t
            return a <= max_chain_length and b <= max_chain_length and abs(a - b) <= chain_imbalance

        def sortedpair(k):
            return min(k), max(k)

        feasible_sizes = {mn for mn, S in Siz2Len.items()
                          if any(map(acceptable_chains, S))}
        m, n = max(feasible_sizes, key=sortedpair)
        best_r = None
        best_ab = overkill, overkill
        for mn in set(((m, n), (n, m))) & feasible_sizes:
            for ab, r in Siz2Len[mn].items():
                ab = max(ab), min(ab)
                if acceptable_chains(ab) and ab < best_ab:
                    best_ab = ab
                    best_r = r

        bestsize = sortedpair(self.biclique_size(*best_r))
        bestbiclique = self.biclique(*best_r)
        return bestsize, (bestbiclique[0], bestbiclique[1])

    def tightestNativeBiClique(self, n, m=None, chain_imbalance=0, max_chain_length=None):
        """Returns a native embedding for the complete bipartite graph :math:`K_{n,m}`
        with the shortest chains possible.  The native embedding of a complete
        bipartite graph is a set of horizontally-aligned qubits connected
        in lines together with an equal-sized set of vertically-aligned qubits
        connected in lines.

        INPUTS:
            n (int): target size for one side of the bipartition

            m (int): target size for the other side of the bipartition, or
            None for the special case :math:`m=n` (default: ``None``)

            chain_imbalance (int): how big of a difference to allow between the
            chain lengths on the two sides of the bipartition.  If
            ``None``, we allow an arbitrary imbalance. (default: ``0``)

            max_chain_length (int): longest chain length to consider or ``None`` if
            chain lengths are allowed to be unbounded.  (default: ``None``)

        OUTPUT:
            score (tuple): the score for the returned biclique (just ``(min(n,m),max(n,m))`` in the
            class :class:`eden_processor`; may differ in subclasses).  If no :math:`K_{n,m}`
            is available on the processor, this is None.

            embedding (tuple): a tuple of two lists containing lists of qubits.
            If `embedding = (A_side, B_side)`, the lists found in `A_side` and
            `B_side` are chains of qubits.
            These lists of qubits are arranged so that

            >>> [zip(chain,chain[1:]) for chain in A_side]

            and

            >>> [zip(chain,chain[1:]) for chain in B_side]

            are lists of valid couplers.  If no :math:`K_{n,m}` is available on
            the processor, this is None.

        """
        if m is None:
            m = n

        self._compute_biclique_sizes()
        Len2Siz = self._biclique_length_to_size
        Siz2Len = self._biclique_size_to_length
        overkill = self.M + self.N + 1
        if max_chain_length is None:
            max_chain_length = overkill
        if chain_imbalance is None:
            chain_imbalance = overkill

        def acceptable_chains(t):
            a, b = t
            return (a <= max_chain_length and b <= max_chain_length and
                    abs(a - b) <= chain_imbalance)

        def acceptable_size(t):
            m0, n0 = t
            return (m0 >= m and n0 >= n) or (m0 >= n and n0 >= m)

        feasible_sizes = {mn for mn, S in Siz2Len.items()
                          if acceptable_size(mn) and any(map(acceptable_chains, S))}

        best_r = None
        best_ab = overkill, overkill
        best_mn = None
        for mn in feasible_sizes:
            for ab, r in Siz2Len[mn].items():
                ab = max(ab), min(ab)
                if acceptable_chains(ab) and ab < best_ab:
                    best_ab = ab
                    best_r = r
                    best_mn = mn
        if best_r is None:
            return None, None
        bestsize = self.biclique_size(*best_r)
        bestbiclique = self.biclique(*best_r)

        best_m, best_n = best_mn
        if m <= best_m and n <= best_n:
            return bestsize,  (bestbiclique[1][:n], bestbiclique[0][:m])
        else:
            return bestsize,  (bestbiclique[0][:n], bestbiclique[1][:m])


class processor:
    """A class representing a subgraph of Chimera :math:`C_{M,N,L}` for the purpose of
    embedding cliques.  The clique embedding algorithm is polynomial if there
    are a fixed number of broken couplers between working qubits, but
    exponential otherwise (provided those broken couplers connect qubits in
    the same unit cell).  This class wraps the :class:`eden_processor` class, where
    the polynomial algorithm is implemented, and we do the exponential work
    here.

    If there are a small number of bad couplers, we simply store a few
    copies with certain qubits deleted.  Otherwise, memory use would grow
    too large, so we are forced to construct each subprocessor on the fly.
    Perfect optimization gradually becomes impossible, so we default to
    examine no more than 64 subprocessors.  This limit can be increased with
    a parameter to :func:`__init__`.
    """

    def __init__(self, edgelist, M=8, N=None, L=4, proc_limit=64, linear=True,
                 random_bundles=False):
        """Constructs a :class:`processor` instance

        INPUTS:
            edgelist: a list of tuples (p,q) where each of p and q is
                a qubit address in chimera coordinates.  These are taken
                to be the working couplers.
            M,N,L: the chimera parameters, :math:`C_{M,N,L}`.  If N is None, set
                N=M.  Defaults: M=8, N=None, L=4
            proc_limit: the maximum number of subprocessors to examine
                (default 64)
            linear: True if the couplers are given as linear indices
                instead of chimera indices.  (default True)
            random_bundles: True if ell bundles should be shuffled; use this
                to generate a large number of embeddings with (perhaps)
                different performance characteristics. (default False)
        """
        if N is None:
            N = M
        self.M, self.N, self.L = M, N, L
        if linear:
            edgelist = [(_to_chimera(M, N, L, p), _to_chimera(M, N, L, q)) for p, q in edgelist]

        self._linear = linear
        self._proc_limit = proc_limit
        self._qubits = set(q for e in edgelist for q in e)
        self._edgelist = edgelist
        self._random_bundles = random_bundles
        self._proc0 = eden_processor(edgelist, M, N, L, random_bundles=random_bundles)
        self._find_evil()
        self._compute_deletions()

    def _compute_all_deletions(self):
        """Returns all minimal edge covers of the set of evil edges.
        """
        minimum_evil = []
        for disabled_qubits in map(set, product(*self._evil)):
            newmin = []
            for s in minimum_evil:
                if s < disabled_qubits:
                    break
                elif disabled_qubits < s:
                    continue
                newmin.append(s)
            else:
                minimum_evil = newmin + [disabled_qubits]
        return minimum_evil

    def _subprocessor(self, disabled_qubits):
        """Create a subprocessor by deleting a set of qubits.  We assume
        this removes all evil edges, and return an :class:`eden_processor`
        instance.
        """
        edgelist = [(p, q) for p, q in self._edgelist if
                    p not in disabled_qubits and
                    q not in disabled_qubits]
        return eden_processor(edgelist, self.M, self.N, self.L, random_bundles=self._random_bundles)

    def _compute_deletions(self):
        """If there are fewer than self._proc_limit possible deletion
        sets, compute all subprocessors obtained by deleting a
        minimal subset of qubits.
        """
        M, N, L, edgelist = self.M, self.N, self.L, self._edgelist
        if 2**len(self._evil) <= self._proc_limit:
            deletions = self._compute_all_deletions()
            self._processors = [self._subprocessor(d) for d in deletions]
        else:
            self._processors = None

    def _random_subprocessor(self):
        """Creates a random subprocessor where there is a coupler between
        every pair of working qubits on opposite sides of the same cell.
        This is guaranteed to be minimal in that adding a qubit back in
        will reintroduce a bad coupler, but not to have minimum size.

        OUTPUT:
            an :class:`eden_processor` instance
        """
        deletion = set()
        for e in self._evil:
            if e[0] in deletion or e[1] in deletion:
                continue
            deletion.add(choice(e))
        return self._subprocessor(deletion)

    def _random_subprocessors(self):
        """Produces an iterator of subprocessors.  If there are fewer than
        self._proc_limit subprocessors to consider (by knocking out a
        minimal subset of working qubits incident to broken couplers),
        we work exhaustively.  Otherwise, we generate a random set of
        ``self._proc_limit`` subprocessors.

        If the total number of possibilities is rather small, then we
        deliberately pick a random minimum subset to avoid coincidences.
        Otherwise, we give up on minimum, satisfy ourselves with minimal,
        and randomly generate subprocessors with :func:`self._random_subprocessor`.

        OUTPUT:
            an iterator of eden_processor instances.
        """
        if self._processors is not None:
            return (p for p in self._processors)
        elif 2**len(self._evil) <= 8 * self._proc_limit:
            deletions = self._compute_all_deletions()
            if len(deletions) > self._proc_limit:
                deletions = sample(deletions, self._proc_limit)
            return (self._subprocessor(d) for d in deletions)
        else:
            return (self._random_subprocessor() for i in range(self._proc_limit))

    def _map_to_processors(self, f, objective):
        """Map a function to a list of processors, and return the output that
        best satisfies a transitive objective function.  The list of
        processors will differ according to the number of evil qubits and
        :func:`_proc_limit`, see details in :func:`self._random_subprocessors`.

        INPUT:
            f (callable): the function to call on each processor

            objective (callable): a function where objective(x,y) is True if x is
                better than y, and False otherwise.  Assumes transitivity!

        OUTPUT:
            best: the object returned by f that maximizes the objective.
        """

        P = self._random_subprocessors()
        best = f(next(P))
        for p in P:
            x = f(p)
            if objective(best, x):
                best = x
        return best[1]

    def _objective_bestscore(self, old, new):
        """An objective function that returns True if new has a better score
        than old, and ``False`` otherwise.

        INPUTS:
            old (tuple): a tuple (score, embedding)

            new (tuple): a tuple (score, embedding)

        """
        (oldscore, oldthing) = old
        (newscore, newthing) = new
        if oldscore is None:
            return True
        if newscore is None:
            return False
        return oldscore < newscore

    def _objective_qubitcount(self, old, new):
        """An objective function that returns True if new uses fewer qubits
        than old, and False otherwise.  This objective function should only be
        used to compare embeddings of the same graph (or at least embeddings of
        graphs with the same number of qubits).

        INPUTS:
            old (tuple): a tuple (score, embedding)

            new (tuple): a tuple (score, embedding)

        """
        (oldscore, oldthing) = old
        (newscore, newthing) = new

        def measure(chains):
            return sum(map(len, chains))

        if oldscore is None:
            return True
        if newscore is None:
            return False
        if len(newthing):
            if not len(oldthing):
                return True
            elif isinstance(newthing, tuple):
                newlengths = sum(map(measure, newthing))
                oldlengths = sum(map(measure, oldthing))
                return newlengths < oldlengths
            else:
                return measure(newthing) < measure(oldthing)
        else:
            return False

    def _find_evil(self):
        """A utility function that computes a list of missing couplers which
        should connect two working qubits in the same cell.  The presence
        of (a nonconstant number of) these breaks the polynomial-time
        claim for our algorithm.  Note: we're only actually hurt by missing
        intercell couplers.
        """
        M, N, L = self.M, self.N, self.L
        proc = self._proc0
        evil = []
        cells = [(x, y) for x in range(M) for y in range(N)]
        spots = [(u, v) for u in range(L) for v in range(L)]
        for x, y in cells:
            for u, v in spots:
                p = (x, y, 0, u)
                q = (x, y, 1, v)
                if p in proc and q in proc and p not in proc[q]:
                    evil.append((p, q))
        self._evil = evil

    def tightestNativeClique(self, n):
        """Returns the native clique embedding with the shortest chains and
        at least ``n`` completely connected chains, or ``None`` on failure.

        INPUTS:
            n: size of clique to return

        OUTPUT:
            clique (list): a list containing lists of qubits, each associated
            to a chain.  These lists of qubits are carefully
            arranged so that

            >>> [zip(chain,chain[1:]) for chain in clique]

            is a list of valid couplers.  If no ``n``-clique is
            available on the processor, this is ``None.``

        Note: this function does not return a uniformly random choice,
        as the nativeCliqueEmbed function only guarantees uniform choice
        of maximum cliques.
        """
        def f(x):
            return x.tightestNativeClique(n)
        objective = self._objective_qubitcount
        return self._translate(self._map_to_processors(f, objective))

    def largestNativeClique(self, max_chain_length=None):
        """Returns the largest native clique embedding we can find on the
        processor, with the shortest chainlength possible (for that clique
        size).  If possible, returns a uniform choice among all largest
        cliques.

        INPUTS:
            max_chain_length (int): longest chain length to consider or ``None`` if chain
            lengths are allowed to be unbounded.  (default: ``None``)

        OUTPUT:
            clique (list): a list containing lists of qubits, each associated to a
            chain.  These lists of qubits are carefully arranged so that

            >>> [zip(chain,chain[1:]) for chain in clique]

            is a list of valid couplers.

        Note: this fails to return a uniform choice if there are broken
        intra-cell couplers between working qubits. (the choice is
        uniform on a particular subprocessor)
        """
        def f(x):
            return x.largestNativeClique(max_chain_length=max_chain_length)
        objective = self._objective_bestscore
        return self._translate(self._map_to_processors(f, objective))

    def nativeCliqueEmbed(self, width):
        """Compute a maximum-sized native clique embedding in an induced
        subgraph of chimera with chainsize ``width+1``.  If possible,
        returns a uniform choice among all largest cliques.

        INPUTS:
            width: width of the squares to search, also `chainlength-1`

        OUTPUT:
            clique: a list containing lists of qubits, each associated
            to a chain.  These lists of qubits are carefully
            arranged so that

            >>> [zip(chain,chain[1:]) for chain in clique]

            is a list of valid couplers.

        Note: this fails to return a uniform choice if there are broken
        intra-cell couplers between working qubits. (the choice is
        uniform on a particular subprocessor)
        """
        def f(x):
            return x.nativeCliqueEmbed(width)
        objective = self._objective_bestscore
        return self._translate(self._map_to_processors(f, objective))

    def largestNativeBiClique(self, chain_imbalance=0, max_chain_length=None):
        """Returns a native embedding for the complete bipartite graph :math:`K_{n,m}`
        for `n <= m`; where `n` is as large as possible and `m` is as large as
        possible subject to `n`.  The native embedding of a complete bipartite
        graph is a set of horizontally-aligned qubits connected in lines
        together with an equal-sized set of vertically-aligned qubits
        connected in lines.

        INPUTS:
            chain_imbalance: how big of a difference to allow between the
            chain lengths on the two sides of the bipartition. If ``None``,
            then we allow an arbitrary imbalance.  (default: ``0``)

            max_chain_length: longest chain length to consider or None if chain
            lengths are allowed to be unbounded.  (default: ``None``)

        OUTPUT:
            embedding (tuple): a tuple of two lists containing lists of qubits.
            If ``embedding = (A_side, B_side)``, the lists found in ``A_side`` and
            ``B_side`` are chains of qubits. These lists of qubits are arranged so that

            >>> [zip(chain,chain[1:]) for chain in A_side]

            and

            >>> [zip(chain,chain[1:]) for chain in B_side]

            are lists of valid couplers.
        """
        def f(x):
            return x.largestNativeBiClique(chain_imbalance=chain_imbalance,
                                           max_chain_length=max_chain_length)
        objective = self._objective_bestscore
        emb = self._map_to_processors(f, objective)
        return self._translate_partitioned(emb)

    def tightestNativeBiClique(self, n, m=None, chain_imbalance=0, max_chain_length=None):
        """Returns a native embedding for the complete bipartite graph :math:`K_{n,m}`
        with the shortest chains possible.  The native embedding of a complete
        bipartite graph is a set of horizontally-aligned qubits connected
        in lines together with an equal-sized set of vertically-aligned qubits
        connected in lines.

        INPUTS:
            n (int): target size for one side of the bipartition

            m (int): target size for the other side of the bipartition, or
            None for the special case ``m=n`` (default: ``None``)

            chain_imbalance (int): how big of a difference to allow between the
            chain lengths on the two sides of the bipartition.  If
            ``None``, we allow an arbitrary imbalance. (default: ``0``)

            max_chain_length (int): longest chain length to consider or ``None`` if
            chain lengths are allowed to be unbounded.  (default: ``None``)

        OUTPUT:
            embedding (tuple): a tuple of two lists containing lists of qubits.
            If ``embedding = (A_side, B_side)``, the lists found in ```A_side``` and
            ``B_side`` are chains of qubits.
            These lists of qubits are arranged so that

            >>> [zip(chain,chain[1:]) for chain in A_side]

            and

            >>> [zip(chain,chain[1:]) for chain in B_side]

            are lists of valid couplers.  If no :math:`K_{n,m}` is available on
            the processor, this is ``None``.
        """
        def f(x):
            return x.tightestNativeBiClique(n, m=m,
                                            chain_imbalance=chain_imbalance,
                                            max_chain_length=max_chain_length)
        objective = self._objective_qubitcount
        emb = self._map_to_processors(f, objective)
        return self._translate_partitioned(emb)

    def _translate_partitioned(self, embedding):
        """Translates a partitioned embedding back to linear coordinates if
        necessary.  This is useful, for example, in biclique embeddings.  A
        partitioned embedding is a tuple of lists of qubits.
        """
        if embedding is None:
            return None
        return tuple(map(self._translate, embedding))

    def _translate(self, embedding):
        "Translates an embedding back to linear coordinates if necessary."
        if embedding is None:
            return None
        if not self._linear:
            return embedding
        return [_bulk_to_linear(self.M, self.N, self.L, chain) for chain in embedding]


def _bulk_to_linear(M, N, L, qubits):
    "Converts a list of chimera coordinates to linear indices."
    return [2 * L * N * x + 2 * L * y + L * u + k for x, y, u, k in qubits]


def _to_linear(M, N, L, q):
    "Converts a qubit in chimera coordinates to its linear index."
    (x, y, u, k) = q
    return 2 * L * N * x + 2 * L * y + L * u + k


def _bulk_to_chimera(M, N, L, qubits):
    "Converts a list of linear indices to chimera coordinates."
    return [(q // N // L // 2, (q // L // 2) % N, (q // L) % 2, q % L) for q in qubits]


def _to_chimera(M, N, L, q):
    "Converts a qubit's linear index to chimera coordinates."
    return (q // N // L // 2, (q // L // 2) % N, (q // L) % 2, q % L)


def _chimera_neighbors(M, N, L, q):
    "Returns a list of neighbors of (x,y,u,k) in a perfect :math:`C_{M,N,L}`"
    (x, y, u, k) = q
    n = [(x, y, 1 - u, l) for l in range(L)]
    if u == 0:
        if x:
            n.append((x - 1, y, u, k))
        if x < M - 1:
            n.append((x + 1, y, u, k))
    else:
        if y:
            n.append((x, y - 1, u, k))
        if y < N - 1:
            n.append((x, y + 1, u, k))
    return n


def random_processor(M, N, L, qubit_yield, num_evil=0):
    """A utility function that generates a random :math:`C_{M,N,L}` missing some
    percentage of its qubits.

    INPUTS:
        M,N,L: the chimera parameters
        qubit_yield: ratio (0 <= qubit_yield <= 1) of #{qubits}/(2*M*N*L)
        num_evil: number of broken in-cell couplers between working qubits

    OUTPUT:
        proc (:class:`processor`): a :class:`processor` instance with a random
            collection of qubits and couplers as specified
    """
    # replacement for lambda in edge filter below that works with bot h
    def edge_filter(pq):
        # we have to unpack the (p,q) edge
        p, q = pq
        return q in qubits and p < q

    qubits = [(x, y, u, k) for x in range(M) for y in range(N) for u in [0, 1] for k in range(L)]
    nqubits = len(qubits)
    qubits = set(sample(qubits, int(nqubits * qubit_yield)))
    edges = ((p, q) for p in qubits for q in _chimera_neighbors(M, N, L, p))
    edges = list(filter(edge_filter, edges))
    possibly_evil_edges = [(p, q) for p, q in edges if p[:2] == q[:2]]
    num_evil = min(num_evil, len(possibly_evil_edges))
    evil_edges = sample(possibly_evil_edges, num_evil)
    return processor(set(edges) - set(evil_edges), M=M, N=N, L=L, linear=False)



# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from collections import abc

import dwave_networkx as dnx
import networkx as nx
import numpy as np
from scipy import optimize, spatial
from math import ceil
import rpack

def p_norm(G, p=2, starting_layout=None, G_distances=None, dim=None, center=None, scale=None, **kwargs):
    """Embeds graph ``G`` in :math:`R^d` with the p-norm and minimizes a 
    Kamada-Kawai-esque objective function to achieve an embedding with low 
    distortion.
    
    This computes a :class:`.Layout` for ``G`` where the graph distance and the 
    p-distance are very close to each other.

    By default, :func:`p_norm` is used to compute the layout for source graphs 
    when :func:`minorminer.layout.find_embedding` is called.

    Args:
        G (NetworkX Graph):
            The graph to compute the layout for.

        p (int, optional, default=2):
            The order of the p-norm to use as a metric.

        starting_layout (dict, optional, default=None)
            A mapping from the vertices of ``G`` to points in :math:`R^d`. If 
            None, :func:`nx.spectral_layout` is used if possible. Otherwise, 
            :func:`nx.random_layout` is used.

        G_distances (dict, optional, default=None):
            A dictionary of dictionaries representing distances from every vertex 
            in ``G`` to every other vertex in ``G``. If None, it is computed.

        dim (int, optional, default=None):
            The desired dimension of the returned layout, :math:`R^{\dim}`. 
            If None, the dimension of ``center`` is used. If ``center`` is None, 
            ``dim`` is set to 2.

        center (tuple, optional, default=None):
            The desired center point of the returned layout. If None, ``center`` 
            is set as the origin in :math:`R^{\dim}` space.
        
        scale (float, optional, default=None):
            The desired scale of the returned layout; i.e. the layout 
            is in :math:`[center - scale, center + scale]^d` space. If None, no 
            scale is set.

    Returns:
        dict: :attr:`.Layout.layout`, a mapping from vertices of ``G`` (keys) to 
        points in :math:`R^d` (values).
    
    Examples:
        This example creates a :class:`.Layout` object for a hexagonal lattice 
        graph, with coordinates computed using :func:`p_norm`.

        >>> import networkx as nx
        >>> import minorminer.layout as mml
        ...
        >>> G = nx.hexagonal_lattice_graph(2,2)
        >>> layout = mml.Layout(G, mml.p_norm, center=(1,1))

        ``layout`` may be passed in directly to :func:`minorminer.layout.find_embedding`. 
        
        Alternatively, :func:`p_norm` may be passed in instead.

        This next example finds an embedding of a hexagonal lattice graph on a 
        Chimera graph, in which the layouts of both the source and target graphs 
        are computed using p_norm.

        >>> import networkx as nx
        >>> import dwave_networkx as dnx
        >>> import minorminer.layout as mml
        ...
        >>> G = nx.hexagonal_lattice_graph(2,2)
        >>> C = dnx.chimera_graph(2,2)
        >>> embedding = mml.find_embedding(G, 
        ...                                C, 
        ...                                layout=(mml.p_norm, mml.p_norm),
        ...                                center=(1,1))

    """
    dim, center = _set_dim_and_center(dim, center)

    # Use the user provided starting_layout, or a spectral_layout if the dimension 
    # is low enough. If neither, use a random_layout.
    if starting_layout:
        pass
    elif dim >= len(G):
        starting_layout = nx.random_layout(G, dim=dim)
    else:
        starting_layout = nx.spectral_layout(G, dim=dim)

    # Make a layout object
    layout = Layout(G, starting_layout, dim=dim)

    # Save on distance calculations by passing them in
    G_distances = _graph_distance_matrix(G, G_distances)

    # Solve the Kamada-Kawai-esque minimization function
    X = optimize.minimize(
        _p_norm_objective,
        layout.layout_array.ravel(),
        method='L-BFGS-B',
        args=(G_distances, dim, p),
        jac=True,
    )

    # Read out the solution to the minimization problem and save layouts
    layout.layout_array = X.x.reshape(len(G), dim)

    # Center and scale the layout
    layout.center = center

    if scale:
        layout.scale = scale

    return layout.layout


def _graph_distance_matrix(G,
                           all_pairs_shortest_path_length=None, 
                           disconnected_distance=None):
    """Compute the distance matrix of G.

    Args:
        G (NetworkX Graph):
            The graph to find the distance matrix of.
        
        all_pairs_shortest_path_length (dict, optional, default=None):
            If None, it is computed by calling :func:`nx.all_pairs_shortest_path_length`.
    
        disconnected_distance (float, optional, default=None):
            A default distance to use when nodes belong to different connected
            components. If None, `len(G)` is used.

    Returns:
        G_distances (Numpy 2d array):
            An array indexed by sorted vertices of G whose i,j value is d_G(i,j).
    
    """
    if all_pairs_shortest_path_length is None:
        all_pairs_shortest_path_length = nx.all_pairs_shortest_path_length(G)

    if disconnected_distance is None:
        disconnected_distance = len(G)

    return np.array(
        [
            [V.get(v, disconnected_distance) for v in sorted(G)]
                 for u, V in all_pairs_shortest_path_length
        ]
    )

def _p_norm_objective(layout_vector, G_distances, dim, p):
    """Compute the sum of differences squared between the p-norm and the graph 
    distance as well as the gradient.

    Args:
        layout (Numpy Array):
            A vector indexed by sorted vertices of G whose values are points in 
            some metric space.

        G_distances (Numpy 2d array):
            An array indexed by sorted vertices of G whose i,j value is d_G(i,j).
        
        dim (int):
            The dimension of the metric space. This will reshape the flattened 
            array passed in to the cost function.
        
        p (int):
            The order of the p-norm to use.

    Returns:
        float: The sum of differences squared between the metric distance and 
        the graph distance.
    
    """
    # Reconstitute the flattened array that scipy.optimize.minimize passed in
    n = len(G_distances)
    layout = layout_vector.reshape(n, dim)

    # Difference between pairs of points in a 3d matrix
    diff = layout[:, np.newaxis, :] - layout[np.newaxis, :, :]

    # A 2d matrix of the distances between points
    dist = np.linalg.norm(diff, ord=p, axis=-1)

    # TODO: Compare this division-by-zero strategy to adding epsilon.
    # A vectorized version of the gradient function
    with np.errstate(divide='ignore', invalid='ignore'):  # handle division by 0
        if p == 1:
            grad = np.einsum(
                'ijk,ij,ijk->ik',
                2*diff,
                dist - G_distances,
                np.nan_to_num(1/np.abs(diff))
            )
        elif p == float("inf"):
            # Note: It may not be faster to do this outside of einsum
            abs_diff = np.abs(diff)
            x_bigger = abs_diff[:, :, 0] > abs_diff[:, :, 1]

            grad = np.einsum(
                'ijk,ij,ijk,ijk->ik',
                2*diff,
                dist - G_distances,
                np.nan_to_num(1/abs_diff),
                np.dstack((x_bigger, np.logical_not(x_bigger)))
            )
        else:
            grad = np.einsum(
                'ijk,ijk,ij,ij->ik',
                2*diff,
                np.abs(diff)**(p-2),
                dist - G_distances,
                np.nan_to_num((1/dist)**(p-1))
            )

    # Return the cost and the gradient
    return np.sum((G_distances - dist)**2), grad.ravel()


def dnx_layout(G, dim=None, center=None, scale=None, **kwargs):
    """The Chimera or Pegasus layout from `dwave_networkx` centered at the origin
    with ``scale`` as a function of the number of rows or columns. Note: As per 
    the implementation of `dnx.*_layout`, if :math:`dim>2`, coordinates beyond 
    the second are 0.

    By default, :func:`dnx_layout` is used to compute the layout for target 
    Chimera or Pegasus graphs when :func:`minorminer.layout.find_embedding` is 
    called.

    Args:
        G (NetworkX Graph):
            The graph to compute the layout for.

        dim (int, optional, default=None):
            The desired dimension of the returned layout, :math:`R^{\dim}`. If 
            None, the dimension of ``center`` is used. If ``center`` is None, 
            ``dim`` is set to 2.

        center (tuple, optional, default=None):
            The desired center point of the returned layout. If None, it is set 
            as the origin in :math:`R^{\dim}` space.

        scale (float, optional, default=None):
            The desired scale of the returned layout; i.e. the layout is in 
            :math:`[center - scale, center + scale]^d` space. If None, it is set 
            as :math:`max(n, m)/2`, where n, m are the number of columns, rows 
            respectively in ``G``.

    Returns:
        dict: :attr:`.Layout.layout`, a mapping from vertices of ``G`` (keys) to 
        points in :math:`R^d` (values).
    
    Examples:
        This example creates a :class:`.Layout` object for a Pegasus graph, with 
        coordinates computed using :func:`dnx_layout`.

        >>> import networkx as nx
        >>> import minorminer.layout as mml
        ...
        >>> P = dnx.pegasus_graph(4)
        >>> layout = mml.Layout(P, mml.dnx_layout, center=(1,1), scale=2)

    """
    graph_data = G.graph

    family = graph_data.get("family")
    if G.graph.get("family") not in ("chimera", "pegasus", "zephyr"):
        raise ValueError(
            "This strategy is only implemented for Chimera, Pegasus"
            " and Zephyr graphs constructed by dwave_networkx`.")

    dim, center = _set_dim_and_center(dim, center)

    if scale is None:
        m, n = graph_data["rows"], graph_data["columns"]
        scale = max(n, m)/2

    dnx_center, dnx_scale = _nx_to_dnx_layout(center, scale)

    if family == "chimera":
        dnx_layout = dnx.chimera_layout(
            G, dim=dim, center=dnx_center, scale=dnx_scale)
    elif family == "pegasus":
        dnx_layout = dnx.pegasus_layout(
            G, dim=dim, center=dnx_center, scale=dnx_scale)
    elif family == "zephyr":
        dnx_layout = dnx.zephyr_layout(
            G, dim=dim, center=dnx_center, scale=dnx_scale)

    layout = Layout(G, dnx_layout)
    return layout.layout


def _nx_to_dnx_layout(center, scale):
    """This function translates a center and a scale from the networkx convention, 
    :math:`[center - scale, center + scale]^{\dim}`, to the `dwave_networkx` 
    convention, :math:`[center, center-scale] x [center, center+scale]^{\(dim-1)}`.

    Returns:
        tuple: A 2-tuple:
            float: The top left corner of a layout.

            float: A scale value that is twice the original scale.
    
    """
    dnx_center = (center[0] - scale, ) + tuple(x + scale for x in center[1:])
    dnx_scale = 2*scale

    return dnx_center, dnx_scale


class Layout(abc.MutableMapping):
    """Class that stores (or computes) coordinates in dimension ``dim`` for each 
    node in graph ``G``.

    Args:
        G (NetworkX Graph/edges data structure (dict, list, ...)):
            The graph to compute the layout for or a NetworkX supported data 
            structure for edges (see :func:`nx.convert.to_networkx_graph` for 
            details).

        layout (dict/function, optional, default=None):
            If a dict, this specifies a pre-computed layout for ``G``.
            
            If a function, this should be in the form of ``layout(G, **kwargs)``,
            in which ``dim``, ``center``, and ``scale`` are passed in as kwargs 
            and the return value is a dict representing a layout of ``G``.
            
            If None, :func:`nx.random_layout(G, **kwargs)` is called.

        dim (int, optional, default=None):
            The desired dimension of the computed layout, :math:`R^{\dim}`. If 
            None, ``dim`` is set as the dimension of ``layout``.

        center (tuple, optional, default=None):
            The desired center point of the computed layout. If None, ``center`` 
            is set as the center of ``layout``.

        scale (float, optional, default=None):
            The desired scale of the computed layout; i.e. the layout is in 
            :math:`[center - scale, center + scale]^d` space. If None, ``scale`` 
            is set to be the scale of ``layout``.

        pack_components (bool, optional, default=True):
            If True, and if the graph contains multiple components and ``dim`` 
            is None or 2, the components will be laid out individually and packed 
            into a rectangle.

        **kwargs (dict):
            Keyword arguments are passed to ``layout`` if it is a function.
    
    """

    def __init__(
        self,
        G,
        layout=None,
        dim=None,
        center=None,
        scale=None,
        pack_components = True,
        **kwargs
    ):
        # Ensure G is a graph object
        self.G = _parse_graph(G)

        # Add dim, center, and scale to kwargs if not None
        if dim is not None:
            kwargs["dim"] = dim
        if center is not None:
            kwargs["center"] = center
        if scale is not None:
            kwargs["scale"] = scale

        _call_layout = False
        # If passed in, save or compute the layout
        if layout is None:
            if self.G.graph.get('family') in ('pegasus', 'chimera', 'zephyr'):
                self.layout = dnx_layout(self.G, **kwargs)
            else:
                _call_layout = True
                layout = p_norm
        elif callable(layout):
            _call_layout = True
        else:
            # Assumes layout implements a mapping interface
            self.layout = layout

        if _call_layout:
            if pack_components:
                self.layout = _pack_components(self.G, layout, **kwargs)
            else:
                self.layout = layout(self.G, **kwargs)

        # Set specs in the order of (user input, precomputed layout)
        self.dim = dim or self._dim
        self.scale = scale or self._scale
        if center is not None:
            self.center = center
        else:
            self.center = self._center

    # Keep layout and layout_array in lockstep with each other.
    @property
    def layout(self):
        return self._layout

    @layout.setter
    def layout(self, value):
        """If layout is set, also set layout_array and the layout specs."""
        # Set the layout
        self._layout = value

        # Iterating through G determines the order of layout_array
        self._layout_array = np.array([value[v] for v in sorted(self.G)])

        self._set_layout_specs()

    @property
    def layout_array(self):
        return self._layout_array

    @layout_array.setter
    def layout_array(self, value):
        """If layout_array is set, also set layout and the layout specs."""
        # Set the layout_array
        self._layout_array = value

        # Update the layout
        self.layout = {v: p for v, p in zip(sorted(self.G), value)}

    @property
    def dim(self):
        return self._dim

    @dim.setter
    def dim(self, value):
        """Set the dimension of the layout, if possible."""
        if value:
            self.layout_array = _dimension_layout(
                self.layout_array, value, self._dim)

    @property
    def center(self):
        return self._center

    @center.setter
    def center(self, value):
        """Recenter the layout."""
        # Cast it as a numpy array incase it's not.
        value = np.array(value)

        if value.size != 0:
            self.layout_array = _center_layout(
                self.layout_array, value, self._center)

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, value):
        """Rescale the layout."""
        if value:
            self.layout_array = _scale_layout(
                self.layout_array, value, self._scale, self._center)

    # The layout class should behave like a dictionary
    def __iter__(self):
        """Iterate through the keys of the dictionary layout."""
        yield from self.layout

    def __getitem__(self, key):
        """Get the layout value at the key vertex."""
        return self.layout[key]

    def __setitem__(self, key, value):
        """Set the layout value at the key vertex."""
        self.layout[key] = value

    def __delitem__(self, key):
        """Delete the layout value at the key vertex."""
        del self.layout[key]

    def __repr__(self):
        """Use the layout's dictionary representation."""
        return repr(self.layout)

    def __len__(self):
        """The length of a layout is the length of the layout dictionary."""
        return len(self.layout)

    def _set_layout_specs(self, empty=False):
        """Set the dimension, center, and scale of the layout_array currently 
        in the Layout object.
        """
        if self.layout_array.size == 0:
            self._dim = 0
            self._center = np.array([])
            self._scale = 0
        else:
            self._dim = self.layout_array.shape[1]
            self._center = _get_center(self.layout_array)
            self._scale = np.max(
                np.linalg.norm(
                    self.layout_array - self._center, float("inf"), axis=0
                )
            )


def _dimension_layout(layout_array, new_d, old_d=None):
    """This helper function transforms a layout from R^old_d to R^new_d by 
    padding extra dimensions with 0's.

    Args:
        layout_array (Numpy array):
            An array whose rows are points in R^old_dim.
        
        new_d (int):
            The new dimension to convert the layout to, must be larger than old_dim.
        
        old_d (int, optional, default=None):
            The current dimension of the layout. If None, the dimension is looked 
            up via layout_array.

    Returns:
        Numpy array: A layout that has been projected into R^new_dim.
    
    """
    # If old_center is empty, compute it
    if old_d is None:
        old_d = layout_array.shape[1]

    if new_d < old_d:
        raise ValueError(
            "The new dimension {} is larger than the old dimension {}. This is not currently supported.".format(
                new_d, old_d)
        )

    n = layout_array.shape[0]
    # Make a block of zeros
    new_layout = np.zeros((n, new_d))
    # Overwrite the entries using layout_array
    new_layout[:, :old_d] = layout_array
    return new_layout


def _center_layout(layout_array, new_center, old_center=None):
    """This helper function transforms a layout from 
    :math:`[old_center - scale, old_center + scale]^d` to
    :math:`[new_center - scale, new_center + scale]^d`.

    Args:
        layout_array (Numpy array):
            An array whose rows are points in R^d.

        new_center (tuple/Numpy array):
            A point in R^d that is the desired center to move the layout to. 
        
        old_center (tuple/Numpy array, optional, default=None):
            A point in R^d representing the center of the layout. If None, the
            approximate center of layout is computed by calculating the center of 
            mass (or centroid).

    Returns:
        Numpy array: A layout that has been centered at new_center.
    
    """
    # If old_center is empty, compute it
    if old_center is None:
        old_center = _get_center(layout_array)

    # Translate the layout so that we have the desired new_center
    return layout_array - old_center + new_center


def _scale_layout(layout_array, new_scale, old_scale=None, center=None):
    """This helper function transforms a layout from 
    :math:`[center - old_scale, center + old_scale]^d` to 
    :math:`[center - new_scale, center + new_scale]^d`.

    Args:
        layout_array (Numpy array):
            An array whose rows are points in R^d.
        
        new_scale (float):
            The desired scale to transform the layout to.
        
        old_scale (float, optional, default=None):
            The scale of the layout. If None, the approximate scale of layout is 
            computed by taking the maximum distance from the center.
        
        center (tuple/Numpy array, optional, default=None):
            A point in R^d representing the center of the layout. If None, the 
            approximate center of layout is computed by calculating the center of 
            mass (centroid).

    Returns:
        Numpy array: A layout that has been scaled to [center - new_scale, center + new_scale]^d.
    
    """
    # If center is empty, compute it
    if center is None:
        center = _get_center(layout_array)

    # Translate the layout so that its center is the origin
    L = layout_array - center

    # Compute the scale of the passed-in layout
    if old_scale is None:
        old_scale = np.max(np.abs(L))

    # Scale the thing
    scaled_L = (new_scale/old_scale) * L

    # Translate the layout back to where it was
    return scaled_L + center


def _set_dim_and_center(dim, center, default_dim=2):
    """A helper function to check that a user provided dim and center match, or 
    if no user provided dim and center exist sets default values: dim=2 and 
    center=the origin in :math:`R^{\dim}`.
    """
    # Set the dimension
    if dim is None:
        if center is None:
            dim = default_dim
        else:
            dim = len(center)

    # Set the center
    if center is None:
        center = dim*(0, )

    if len(center) != dim:
        raise ValueError(
            "Your dimension is {} but your center is {}.".format(dim, center))

    return dim, center


def _rotate_to_minimize_area(points):
    """Uses the rotating calipers algorithm to rotate points (a numpy array)
    into a minimum-area bounding box
    """
    # Compute the convex hull
    hull = spatial.ConvexHull(points, qhull_options = 'QJ')

    # Look up the points that constitute the convex hull
    hull_points = points[hull.vertices]

    pi2 = np.pi/2

    # calculate edge angles
    edges = np.zeros((len(hull_points)-1, 2))
    edges = hull_points[1:] - hull_points[:-1]

    angles = np.zeros((len(edges)))
    angles = np.arctan2(edges[:, 1], edges[:, 0])

    angles = np.abs(np.mod(angles, pi2))
    angles = np.unique(angles)

    # find rotation matrices
    rotations = np.vstack([
        np.cos(angles),
        -np.sin(angles),
        np.sin(angles),
        np.cos(angles)]).T
    rotations = rotations.reshape((-1, 2, 2))

    # apply rotations to the hull
    rot_points = np.dot(rotations, hull_points.T)

    # find the bounding points
    min_x = np.nanmin(rot_points[:, 0], axis=1)
    max_x = np.nanmax(rot_points[:, 0], axis=1)
    min_y = np.nanmin(rot_points[:, 1], axis=1)
    max_y = np.nanmax(rot_points[:, 1], axis=1)

    # find the box with the best area
    widths = (max_x - min_x)
    heights = (max_y - min_y)
    areas =  widths * heights
    best_idx = np.argmin(areas)

    # return the best rotation and its dimensions
    points = np.dot(points, rotations[best_idx])
    min_y = np.nanmin(points[:, 0])
    max_y = np.nanmax(points[:, 0])
    min_x = np.nanmin(points[:, 1])
    max_x = np.nanmax(points[:, 1])
    center = np.array(((max_y + min_y), (max_x + min_x)))/2
    dims = np.array(((max_y - min_y), (max_x - min_x)))
    return points - center, dims

def _pack_components(G, layout, **kwargs):
    """Attempts to pack components using `layout` to compute component layouts,
    rotates those component layouts to minimum-area bounding rectangles,
    and then uses a rectangle packing algorithm to minimize the area of the
    overall layout.

    If `scale` is an argument, we rescale the layout after the above.

    If `dim` is an argument and not equal to 2, we don't have an
    implementation, so we pass the graph directly into `layout` and bypass
    the packing procedure
    """
    if kwargs.get('dim', 2) != 2:
        return layout(G, **kwargs)

    layouts = []
    rectangles = []
    if 'scale' in kwargs:
        subkwargs = dict(kwargs)
        scale = subkwargs['scale']
        del subkwargs['scale']
    else:
        subkwargs = kwargs
        scale = None

    for i, c in enumerate(nx.connected_components(G)):
        if len(c) > 2:
            cpos = layout(G.subgraph(c),
                          scale = len(c)**.5,
                          **subkwargs)
            apos = np.array(list(cpos.values()))
            rpos, dims = _rotate_to_minimize_area(apos)
            cpos = dict(zip(cpos, rpos))
            dims = [int(ceil(z)+.001) for z in dims]
        elif len(c) == 2:
            cpos = dict(zip(c, [(-.5, 0), (.5, 0)]))
            dims = [2, 1]
        else:
            cpos = dict(zip(c, [(0, 0)]))
            dims = [1, 1]
        layouts.append(cpos)
        rectangles.append(dims)

    positions = rpack.pack(rectangles)
    layout = {
        v: (vx + rx + w/2, vy + ry + h/2)
        for (rx, ry), (w, h), cpos in zip(positions, rectangles, layouts)
        for v, (vx, vy) in cpos.items()
    }
    if scale is not None:
        layout_array = np.array(list(layout[v] for v in G))
        return dict(zip(G, _scale_layout(layout_array, scale)))
    else:
        return layout

def _parse_graph(G):
    """Checks that G is a nx.Graph object or tries to make one."""
    if hasattr(G, "edges"):
        return G
    return nx.Graph(G)

def _get_center(layout_array):
    """Compute the center of `layout_array`"""
    mins = np.min(layout_array, axis=0)
    maxs = np.max(layout_array, axis=0)
    return (mins + maxs)/2
